from .features import Feature, FeatureSelector
from .. import (weapons, armor)
from .rogue import UncannyDodge, Evasion


# PHB
class FavoredEnemy(Feature):
    """Beginning at 1st level, you have significant experience studying, tracking,
    hunting, and even talking to a certain type of enemy.

    Choose a type of favored enemy: aberrations, beasts, celestials,
    constructs, dragons, elementals, fey, fiends, giants, monstrosities, oozes,
    plants, or undead. Alternatively, you can select two races of humanoid
    (such as gnolls and orcs) as favored enemies. You have advantage on Wisdom
    (Survival) checks to track your favored enemies, as well as on Intelligence
    checks to recall information about them.

    When you gain this feature, you also learn one language of your choice that
    is spoken by your favored enemies, if they speak one at all. You choose one
    additional favored enemy, as well as an associated language, at 6th and
    14th level. As you gain levels, your choices should reflect the types of
    monsters you have encountered on your adventures.

    """
    name = "Favored Enemy"
    source = "Ranger"
    languages = ("[Select One]",)


class NaturalExplorer(Feature):
    """You are particularly familiar with one type of natural environment and are
    adept at traveling and surviving in such regions. Choose one type of
    favored terrain: arctic, coast, desert, forest, grassland, mountain, swamp,
    or the Underdark.  You choose additional favored terrain types at 6th and
    10th

    When you make an Intelligence or Wisdom check related to your favored
    terrain, your proficiency bonus is doubled if you are using a skill that
    you’re proficient in. While traveling for an hour or more in your favored
    terrain, you gain the following benefits:

    -- Difficult terrain doesn’t slow your group’s travel.

    -- Your group can’t become lost except by magical means.
    
    -- Even when you are engaged in another activity while traveling
    (such as foraging, navigating, or tracking), you remain alert to danger. 

    -- If you are traveling alone, you can move stealthily at a normal pace.

    -- When you forage, you find twice as much food as you normally would.

    -- While tracking other creatures, you also learn their exact number, their
    sizes, and how long ago they passed through the area.

    """
    name = "Natural Explorer"
    source = "Ranger"


class Archery(Feature):
    """
    You gain a +2 bonus to attack rolls you make
    with ranged weapons (included in stats on Character Sheet).
    """
    name = "Fighting Style (Archery)"
    source = "Ranger"

    def weapon_func(self, weapon: weapons.Weapon, **kwargs):
        """
        +2 attack roll bonus if weapon is ranged
        """
        if isinstance(weapon, weapons.RangedWeapon):
            weapon.attack_bonus += 2

    
class Defense(Feature):
    """
    While you are wearing armor, you gain a +1 bonus to AC (included in 
    stats on Character Sheet).
    """
    name = "Fighting Style (Defense)"
    source = "Ranger"

        
class Dueling(Feature):
    """When you are wielding a melee weapon in one hand and no other weapons, you
    gain a +2 bonus to damage rolls with that weapon.

    """
    name = "Fighting Style (Dueling)"
    source = "Ranger"

    def weapon_func(self, weapon: weapons.Weapon, **kwargs):
        """
        +2 attack roll bonus if melee weapon is not two handed
        """
        if (isinstance(weapon, weapons.MeleeWeapon)
                and "two-handed" not in weapon.properties.lower()):
            weapon.damage_bonus += 2

    
class TwoWeaponFighting(Feature):
    """When you engage in two-weapon fighting, you can add your ability modifier
    to the damage of the second attack.

    """
    name = "Fighting Style (Two-Weapon Fighting)"
    source = "Ranger"


class RangerFightingStyle(FeatureSelector):
    """
    Select a Fighting Style by choosing in feature_choices:

    archery
 
    defense

    dueling

    two-weapon fighting
    """
    options = {'archery': Archery,
               'defense': Defense,
               'dueling': Dueling,
               'two-weapon fighting': TwoWeaponFighting,
               'two-weapon': TwoWeaponFighting,
               'dual wield': TwoWeaponFighting}
    name = "Fighting Style (Select One)"
    source = "Ranger"


class PrimevalAwareness(Feature):
    """Beginning at 3rd level, you can use your action and expend one ranger spell
    slot to focus your awareness on the region around you. For 1 minute per
    level of the spell slot you expend, you can sense whether the following
    types of creatures are present within 1 mile of you (or within up to 6
    miles if you are in your favored terrain): aberrations, celestials,
    dragons, elementals, fey, fiends, and undead. This feature doesn’t reveal
    the creatures’ location or number.

    """
    name = "Primeval Awareness"
    source = "Ranger"


class ExtraAttackRanger(Feature):
    """Beginning at 5th level, you can attack twice, instead of once, whenever you
    take the Attack action on your turn.

    """
    name = "Extra Attack (2x)"
    source = "Ranger"


class HideInPlainSight(Feature):
    """Starting at 10th level, you can spend 1 minute creating camouflage for
    yourself. You must have access to fresh mud, dirt, plants, soot, and other
    naturally occurring materials with which to create your camouflage. Once
    you are camouflaged in this way, you can try to hide by pressing yourself
    up against a solid surface, such as a tree or wall, that is at least as
    tall and wide as you are.

    You gain a +10 bonus to Dexterity (Stealth) checks as long as you remain
    there without moving or taking actions. Once you move or take an action or
    a reaction, you must camouflage yourself again to gain this benefit

    """
    name = "Hide in Plain Sight"
    source = "Ranger"


class Vanish(Feature):
    """Starting at 14th level, you can use the Hide action as a bonus action on
    your turn. Also, you can’t be tracked by nonmagical means, unless you
    choose to leave a trail.

    """
    name = "Vanish"
    source = "Ranger"


class FeralSenses(Feature):
    """At 18th level, you gain preternatural senses that help you fight creatures
    you can’t see. When you attack a creature you can’t see, your inability to
    see it doesn’t impose disadvantage on your attack rolls against it. You are
    also aware of the location of any invisible creature within 30 feet of you,
    provided that the creature isn’t hidden from you and you aren’t blinded or
    deafened

    """
    name = "Feral Senses"
    source = "Ranger"


class FoeSlayer(Feature):
    """At 20th level, you become an unparalleled hunter of your enemies. Once on
    each of your turns, you can add your Wisdom modifier to the attack roll or
    the damage roll of an attack you make against one o f your favored
    enemies. You can choose to use this feature before or after the roll, but
    before any effects of the roll are applied.

    """
    name = "Foe Slayer"
    source = "Ranger"

    
# Hunter
class ColossusSlayer(Feature):
    """Your tenacity can wear down the most potent foes. When you hit a creature
    with a weapon attack, the creature takes an extra 1d8 damage if it’s below
    its hit point maximum. You can deal this extra damage only once per turn.

    """
    name = "Colossus Slayer"
    source = "Ranger (Hunter)"


class GiantKiller(Feature):
    """When a Large or larger creature within 5 feet of you hits or misses you
    with an attack, you can use your reaction to attack that creature
    immediately after its attack, provided that you can see the creature.

    """
    name = "Giant Killer"
    source = "Ranger (Hunter)"


class HordeBreaker(Feature):
    """Once on each o f your turns when you make a weapon attack, you can make
    another attack with the same weapon against a different creature that is
    within 5 feet of the original target and within range of your weapon.

    """
    name = "Horde Breaker"
    source = "Ranger (Hunter)"


class HuntersPrey(FeatureSelector):
    """Select a Hunter's Prey option in "feature_choices" in your .py file from
    one of:

    colossus slayer

    giant killer

    horde breaker

    """
    options = {'colossus slayer': ColossusSlayer,
               'giant killer': GiantKiller,
               'horde breaker': HordeBreaker}
    name = "Hunter's Prey (Select One)"
    source = "Ranger (Hunter)"
    

class EscapeTheHorde(Feature):
    """Opportunity attacks against you are made with disadvantage

    """
    name = "Escape the Horde"
    source = "Ranger (Hunter)"


class MultiattackDefense(Feature):
    """When a creature hits you with an attack, you gain a +4 bonus to AC against
    all subsequent attacks made by that creature for the rest of the turn.

    """
    name = "Multiattack Defense"
    source = "Ranger (Hunter)"


class SteelWill(Feature):
    """You have advantage on saving throws against being frightened.

    """
    name = "Steel Will"
    source = "Ranger (Hunter)"


class DefensiveTactics(FeatureSelector):
    """Select a Defensive Tactics option in "feature_choices" in your .py file from
    one of:

    escape the horde

    multiattack defense

    steel will

    """
    options = {'escape the horde': EscapeTheHorde,
               'multiattack defense': MultiattackDefense,
               'steel will': SteelWill}
    name = "Defensive Tactics (Select One)"
    source = "Ranger (Hunter)"


class Volley(Feature):
    """You can use your action to make a ranged attack against any number of
    creatures within 10 feet o f a point you can see within your weapon’s
    range. You must have ammunition for each target, as normal, and you make a
    separate attack roll for each target

    """
    name = "Volley"
    source = "Ranger (Hunter)"


class WhirlwindAttack(Feature):
    """You can use your action to make a melee attack against any number o f
    creatures within 5 feet of you, with a separate attack roll for each target

    """
    name = "Whirlwind Attack"
    source = "Ranger (Hunter)"


class MultiattackRanger(FeatureSelector):
    """Select a Multiattack option in "feature_choices" in your .py file from
    one of:

    volley

    whirlwind attack

    """
    options = {'volley': Volley,
               'whirlwind attack': WhirlwindAttack}
    name = "Multiattack (Select One)"
    source = "Ranger (Hunter)"


class StandAgainstTheTide(Feature):
    """When a hostile creature misses you with a melee attack, you can use your
    reaction to force that creature to repeat the same attack against another
    creature (other than itself) of your choice

    """
    name = "Stand Against the Tide"
    source = "Ranger (Hunter)"


class SuperiorHuntersDefense(FeatureSelector):
    """Select a Superior Hunter's Defense option in "feature_choices" in your .py
    file from one of:

    evasion
    
    stand against the tide

    uncanny dodge

    """
    options = {'evasion': Evasion,
               'stand against of the tide': StandAgainstTheTide,
               'uncanny dodge': UncannyDodge}
    name = "Superior Hunter's Defense (Select One)"
    source = "Ranger (Hunter)"


# Beast Master
class RangersCompanion(Feature):
    """At 3rd level, you gain a beast companion that accompanies you on your
    adventures and is trained to fight alongside you. Choose a beast that is no
    larger than Medium and that has a challenge rating of 1/4 or lower
    (appendix D presents statistics for the hawk, mastiff, and panther as
    examples). Add your proficiency bonus to the beast’s AC, attack rolls, and
    damage rolls, as well as to any saving throws and skills it is proficient
    in. Its hit point maximum equals its normal maximum or four times your
    ranger level, whichever is higher.

    The beast obeys your commands as best as it can. It takes its turn on your
    initiative, though it doesn’t take an action unless you command it to. On
    your turn, you can verbally command the beast where to move (no action
    required by you). You can use your action to verbally command it to take
    the Attack, Dash, Disengage, Dodge, or Help action. Once you have the Extra
    Attack feature, you can make one weapon attack yourself when you command
    the beast to take the Attack action.

    While traveling through your favored terrain with only the beast, you can
    move stealthily at a normal pace. If the beast dies, you can obtain another
    one by spending 8 hours magically bonding with another beast that isn’t
    hostile to you, either the same type of beast as before or a different one.

    """
    name = "Ranger's Companion"
    source = "Ranger (Beast Master)"


class ExceptionalTraining(Feature):
    """Beginning at 7th level, on any of your turns when your beast companion
    doesn’t attack, you can use a bonus action to command the beast to take the
    Dash, Disengage, Dodge, or Help action on its turn

    """
    name = "Exceptional Training"
    source = "Ranger (Beast Master)"


class BestialFury(Feature):
    """Starting at 11th level, your beast companion can make two attacks when you
    command it to use the Attack action.

    """
    name = "Bestial Fury"
    source = "Ranger (Beast Master)"


class ShareSpells(Feature):
    """Beginning at 15th level, when you cast a spell targeting yourself, you can
    also affect your beast companion with the spell if the beast is within 30
    feet of you

    """
    name = "Share Spells"
    source = "Ranger (Beast Master)"


# Gloom Stalker
class DreadAmbusher(Feature):
    """At 3rd level, you master the art of the ambush. You can give yourself a
    bonus to your initiative rolls equal to your Wisdom modifier. At the start
    of your first turn of each combat, your walking speed increases by 10 feet,
    which lasts until the end of that turn. If you take the Attack action on
    that turn, you can make one additional weapon attack as part of that
    action. If that attack hits, the target takes an extra 1d8 damage of the
    weapons damage type.

    """
    name = 'Dread Ambusher'
    source = "Ranger (Gloom Stalker)"


class UmbralSight(Feature):
    """At 3rd level, you gain darkvision out to a range of 60 feet. If you already
    have darkvision from your race, its range increases by 30 feet. You are
    also adept at evading creatures that rely on darkvision. While in darkness,
    you are invisible to any creature that relies on darkvision to see you in
    that darkness.

    """
    name = "Umbral Sight"
    source = "Ranger (Gloom Stalker)"


class IronMind(Feature):
    """By 7th level, you have honed your ability to resist the mind—altering
    powers of your prey. You gain proficiency in Wisdom saving throws. Ifyou
    already have this proficiency, you instead gain proficiency in
    Intelligence or Charisma saving throws (your choice)

    """
    name = "Iron Mind"
    source = "Ranger (Gloom Stalker)"
    needs_implementation = True

    def __init__(self, owner=None):
        super().__init__(owner=owner)


class StalkersFlurry(Feature):
    """At 11th level, you learn to attack with such unexpected speed that you can
    turn a miss into another strike. Once on each of your turns when you miss
    with a weapon attack, you can make another weapon attack as part of the
    same action

    """
    name = "Stalker's Flurry"
    source = "Ranger (Gloom Stalker)"


class ShadowyDodge(Feature):
    """Starting at 15th level, you can dodge in unforeseen ways, with wisps of
    supernatural shadow around you. Whenever a creature makes an attack roll
    against you and doesn’t have advantage on the roll, you can use your
    reaction to impose disadvantage on it. You must use this feature before you
    know the outcome of the attack roll.

    """
    name = "Shadowy Dodge"
    source = "Ranger (Gloom Stalker)"


# Horizon Walker
class DetectPortal(Feature):
    """At 3rd level, you gain the ability to magically sense the presence of a
    planar portal. As an action, you detect the distance and direction to the
    closest planar portal within 1 mile of you. Once you use this feature, you
    can’t use it again until you finish a short or long rest.

    """
    name = "Detect Portal"
    source = "Ranger (Horizon Walker)"


class PlanarWarrior(Feature):
    """At 3rd level, you learn to draw on the energy of the multiverse to
    augment your attacks. As a bonus action, choose one creature you can see
    within 30 feet of you. The next time you hit that creature on this turn
    with a weapon attack, all damage dealt by the attack becomes force damage,
    and the creature takes an extra 1d8 force damage from the attack. When you
    reach 11th level in this class, the extra damage increases to 2d8.

    """
    _name = "Planar Warrior"
    source = "Ranger (Horizon Walker)"

    @property
    def name(self):
        if self.owner.Ranger.level < 11:
            return self._name + " (1d8/f)"
        else:
            return self._name + " (2d8/f)"

        
class EtherealStep(Feature):
    """At 7th level, you learn to step through the Ethereal Plane. As a bonus
    action, you can cast the etherealncss spell with this feature, without
    expending a spell slot, but the spell ends at the end of the current
    turn. Once you use this feature, you can’t use it again until you finish a
    short or long rest

    """
    name = "Ethereal Step"
    source = "Ranger (Horizon Walker)"


class DistantStrike(Feature):
    """At 11th level, you gain the ability to pass between the planes in the blink
    of an eye. When you take the Attack action, you can teleport up to 10 feet
    before each attack to an unoccupied space you can see. Ifyou attack at
    least two different creatures with the action, you can make one additional
    attack with it against a third creature.

    """
    name = "Distant Strike"
    source = "Ranger (Horizon Walker)"


class SpectralDefense(Feature):
    """At 15th level, your ability to move between planes enables you to slip
    through the planar boundaries to lessen the harm done to you during
    battle. When you take damage from an attack, you can use your reaction to
    give yourself resistance to all of that attack’s damage on this turn

    """
    name = "Spectral Defense"
    source = "Ranger (Horizon Walker)"


# Monster Slayer
class HuntersSense(Feature):
    """At 3rd level, you gain the ability to peer at a creature and magically
    discern how best to hurt it. As an action, choose one creature you can see
    within 60 feet ofyou. You immediately learn whether the creature has any
    damage immunities, resistances, or vulnerabilities and What they are. If
    the creature is hidden from divination magic, you sense that it has no
    damage immunities, re— sistances, or vulnerabilities. You can use this
    feature a number of times equal to your Wisdom modifier (minimum of
    once). You regain all expended uses of it when you finish a long rest.

    """
    _name = "Hunter's Sense"
    source = "Ranger (Monster Slayer)"

    @property
    def name(self):
        num = max(1, self.owner.wisdom.modifier)
        return self._num + " ({:d}x/LR)".format(num)


class SlayersPrey(Feature):
    """Starting at 3rd level, you can focus your ire on one foe, increasing the
    harm you inflict on it. As a bonus action, you designate one creature you
    can see within 60 feet of you as the target of this feature. The first time
    each turn that you hit that target with a weapon attack, it takes an extra
    1d6 damage from the weapon. This benefit lasts until you finish a short or
    long rest. It ends early if you designate a different creature

    """
    name = "Slayer's Prey"
    source = "Ranger (Monster Slayer)"

    
class SupernaturalDefense(Feature):
    """At 7th level, you gain extra resilience against your prey’s assaults on
    your mind and body. Whenever the target of your Slayer’s Prey forces you to
    make a saving throw and whenever you make an ability check to escape that
    targets grapple, add 1d6 to your roll

    """
    name = "Supernatural Defense"
    source = "Ranger (Monster Slayer)"


class MagicUsersNemesis(Feature):
    """At 11th level, you gain the ability to thwart someone else’s magic. When
    you see a creature casting a spell or teleporting within 60 feet of you,
    you can use your reaction to try to magically foil it. The creature must
    succeed on a Wisdom saving throw against your spell save DC, or its spell
    or teleport fails and is wasted. Once you use this feature, you can’t use
    it again until you finish a short or long rest.

    """
    name = "Magic User's Nemesis"
    source = "Ranger (Monster Slayer)"


class SlayersCounter(Feature):
    """At 15th level, you gain the ability to counterattack when your prey tries
    to sabotage you. If the target of your Slayer’s Prey forces you to make a
    saving throw, you can use your reaction to make one weapon attack against
    the quarry. You make this attack immediately before making the saving
    throw. If your attack hits, your save automatir cally succeeds, in addition
    to the attack’s normal effects

    """
    name = "Slayer's Counter"
    source = "Ranger (Monster Slayer)"
    

# Revised Ranger
class FavoredEnemyRevised(Feature):
    """Beginning at 1st level, you have significant experience studying, tracking,
    hunting, and even talking to a certain type of enemy commonly encountered
    in the wilds.

    Choose a type of favored enemy: beasts, fey, humanoids, monstrosities, or
    undead. You gain a +2 bonus to damage rolls with weapon attacks against
    creatures of the chosen type. Additionally, you have advantage on Wisdom
    (Survival) checks to track your favored enemies, as well as on Intelligence
    checks to recall information about them.

    When you gain this feature, you also learn one language of your choice,
    typically one spoken by your favored enemy or creatures associated with
    it. However, you are free to pick any language you wish to learn

    """
    name = "Favored Enemy"
    source = "Revised Ranger"


class NaturalExplorerRevised(Feature):
    """You are a master of navigating the natural world, and you react with swift
    and decisive action when attacked. This grants you the following benefits:

    --You ignore difficult terrain.

    --You have advantage on initiative rolls.

    --On your first turn during combat, you have advantage on attack rolls
    against creatures that have not yet acted.

    In addition, you are skilled at navigating the wilderness. You gain the
    following benefits when traveling for an hour or more:

    --Difficult terrain doesn’t slow your group’s travel.

    --Your group can’t become lost except by magical means.
    
    --Even when you are engaged in another activity while traveling (such as
    foraging, navigating, or tracking), you remain alert to danger.

    --If you are traveling alone, you can move stealthily at a normal pace.

    --When you forage, you find twice as much food as you normally would.
    
    --While tracking other creatures, you also learn their exact number, their
    sizes, and how long ago they passed through the area.

    """
    name = "Natural Explorer"
    source = "Revised Ranger"

    
class PrimevalAwarenessRevised(Feature):
    """Beginning at 3rd level, your mastery of ranger lore allows you to establish
    a powerful link to beasts and to the land around you.

    You have an innate ability to communicate with beasts, and they recognize
    you as a kindred spirit. Through sounds and gestures, you can communicate
    simple ideas to a beast as an action, and can read its basic mood and
    intent. You learn its emotional state, whether it is affected by magic of
    any sort, its short-term needs (such as food or safety), and actions you
    can take (if any) to persuade it to not attack.

    You cannot use this ability against a creature that you have attacked
    within the past 10 minutes.

    Additionally, you can attune your senses to determine if any of your
    favored enemies lurk nearby. By spending 1 uninterrupted minute in
    concentration (as if you were concentrating on a spell), you can sense
    whether any of your favored enemies are present within 5 miles of you. This
    feature reveals which of your favored enemies are present, their numbers,
    and the creatures’ general direction and distance (in miles) from you.

    If there are multiple groups of your favored enemies within range, you
    learn this information for each group.

    """
    name = "Primeval Awareness"
    source = "Revised Ranger"


class GreaterFavoredEnemy(Feature):
    """At 6th level, you are ready to hunt even deadlier game. Choose a type of
    greater favored enemy: aberrations, celestials, constructs, dragons,
    elementals, fiends, or giants. You gain all the benefits against this
    chosen enemy that you normally gain against your favored enemy, including
    an additional language. Your bonus to damage rolls against all your favored
    enemies increases to +4.

    Additionally, you have advantage on saving throws against the spells and
    abilities used by a greater favored enemy.

    """
    name = "Greated Favored Enemy"
    source = "Revised Ranger"

    
class FleetOfFoot(Feature):
    """Beginning at 8th level, you can use the Dash action as a bonus action on
    your turn.

    """
    name = "Fleet of Foot"
    source = "Revised Ranger"


class HideInPlainSightRevised(Feature):
    """Starting at 10th level, you can remain perfectly still for long periods of
    time to set up ambushes.

    When you attempt to hide on your turn, you can opt to not move on that
    turn. If you avoid moving, creatures that attempt to detect you take a −10
    penalty to their Wisdom (Perception) checks until the start of your next
    turn. You lose this benefit if you move or fall prone, either 4 voluntarily
    or because of some external effect. You are still automatically detected if
    any effect or action causes you to no longer be hidden.

    If you are still hidden on your next turn, you can continue to remain
    motionless and gain this benefit until you are detected

    """
    name = "Hide in Plain Sight"
    source = "Revised Ranger"


# Beast Conclave
class AnimalCompanion(Feature):
    """At 3rd level, you learn to use your magic to create a powerful bond with a
    creature of the natural world.

    With 8 hours of work and the expenditure of 50 gp worth of rare herbs and
    fine food, you call forth an animal from the wilderness to serve as your
    faithful companion. You normally select you companion from among the
    following animals: an ape, a black bear, a boar, a giant badger, a giant
    weasel, a mule, a panther, or a wolf. However, your DM might pick one of
    these animals for you, based on the surrounding terrain and on what types
    of creatures would logically be present in the area.

    At the end of the 8 hours, your animal companion appears and gains all the
    benefits of your Companion’s Bond ability. You can have only one animal
    companion at a time.

    If your animal companion is ever slain, the magical bond you share allows
    you to return it to life. With 8 hours of work and the expenditure of 25 gp
    worth of rare herbs and fine food, you call forth your companion’s spirit
    and use your magic to create a new body for it. You can return an animal
    companion to life in this manner even if you do not possess any part of its
    body.

    If you use this ability to return a former animal companion to life while
    you have a current animal companion, your current companion leaves you and
    is replaced by the restored companion

    """
    name = "Animal Companion"
    source = "Revised Ranger (Animal Companion)"


class CompanionsBond(Feature):
    """Your animal companion gains a variety of benefits while it is linked to
    you. The animal companion loses its Multiattack action, if it has one.

    The companion obeys your commands as best it can. It rolls for initiative
    like any other creature, but you determine its actions, decisions,
    attitudes, and so on. If you are incapacitated or absent, your companion
    acts on its own. When using your Natural Explorer feature, you and your
    animal companion can both move stealthily at a normal pace.

    Your animal companion has abilities and game statistics determined in part
    by your level. Your companion uses your proficiency bonus rather than its
    own. In addition to the areas where it normally uses its proficiency bonus,
    an animal companion also adds its proficiency bonus to its AC and to its
    damage rolls.

    Your animal companion gains proficiency in two skills of your choice. It
    also becomes proficient with all saving throws. For each level you gain
    after 3rd, your animal companion gains an additional hit die and increases
    its hit points accordingly. Whenever you gain the Ability Score Improvement
    class feature, your companion’s abilities also improve. Your companion can
    increase one ability score of your choice by 2, or it can increase two
    ability scores of your choice by 1. As normal, your companion can’t
    increase an ability score above 20 using this feature unless its
    description specifies otherwise.

    Your companion shares your alignment, and has a personality trait and a
    flaw that you can roll for or select from the tables below. Your companion
    shares your ideal, and its bond is always, \“The ranger who travels with me
    is a beloved companion for whom I would gladly give my life.\”

    """
    name = "Companions Bond"
    source = "Revised Ranger (Beast Conclave)"


class CoordinatedAttack(Feature):
    """Beginning at 5th level, you and your animal companion form a more potent
    fighting team. When you use the Attack action on your turn, if your
    companion can see you, it can use its reaction to make a melee attack.
    """
    name = "Coordinated Attack"
    source = "Revised Ranger (Beast Conclave)"


class BeastsDefense(Feature):
    """At 7th level, while your companion can see you, it has advantage on all
     saving throw

    """
    name = "Beast's Defense"
    source = "Revised Ranger (Beast Conclave)"


class StormOfClawsAndFangs(Feature):
    """At 11th level, your companion can use its action to make a melee attack
    against each creature of its choice within 5 feet of it, with a separate
    attack roll for each target

    """
    name = "Storm of Claws and Fangs"
    source = "Revised Ranger (Beast Conclave)"


class SuperiorBeastsDefense(Feature):
    """At 15th level, whenever an attacker that your companion can see hits it
     with an attack, it can use its reaction to halve the attack’s damage
     against it.

    """
    name = "Superior Beast's Defense"
    source = "Revised Ranger (Beast Conclave)"


# Deep Stalker Conclave
class UnderdarkScout(Feature):
    """At 3rd level, you master the art of the ambush. On your first turn during
    combat, you gain a +10 bonus to your speed, and if you use the Attack
    action, you can make one additional attack. You are also adept at evading
    creatures that rely on darkvision. Such creatures gain no benefit when
    attempting to detect you in dark and dim conditions. Additionally, when the
    DM determines if you can hide from a creature, that creature gains no
    benefit from its darkvision

    """
    name = "Underdark Scout"
    source = "Revised Ranger (Deep Stalker Conclave)"


class StalkersDodge(Feature):
    """At 15th level, whenever a creature attacks you and does not have advantage,
    you can use your reaction to impose disadvantage on the creature’s attack
    roll against you. You can use this feature before or after the attack roll
    is made, but it must be used before the outcome of the roll is determined

    """
    name = "Stalker's Dodge"
    source = "Revised Ranger (Deep Stalker Conclave)"




class Weapon():
    name = ""
    cost = "0 gp"
    base_damage = "1d4"
    damage_bonus = 0
    attack_bonus = 0
    damage_type = "p"
    weight = 1  # In lbs
    properties = ""
    ability = 'strength'
    is_finesse = False
    features_applied = False

    def __init__(self, wielder=None):
        self.wielder = wielder

    @classmethod
    def improved_version(cls, bonus):
        bonus = int(bonus)
        
        class NewWeapon(cls):
            name = f'+{bonus} ' + cls.name
            damage_bonus = bonus
            attack_bonus = bonus
            
        return NewWeapon

    def apply_features(self):
        if (not self.features_applied) and (self.wielder is not None):
            self.features_applied = True
            for f in self.wielder.features:
                f.weapon_func(self)

    @property
    def ability_mod(self):
        if self.wielder is None:
            return 0
        else:
            if self.is_finesse:
                return max(self.wielder.strength.modifier,
                           self.wielder.dexterity.modifier)
            else:
                return getattr(self.wielder, self.ability).modifier
        
    @property
    def attack_modifier(self):
        self.apply_features()
        mod = self.attack_bonus
        if self.wielder is not None:
            mod += self.ability_mod
            if self.wielder.is_proficient(self):
                mod += self.wielder.proficiency_bonus
        return mod
    
    @property
    def damage(self):
        self.apply_features()
        dam_str = str(self.base_damage)
        bonus = self.damage_bonus
        if self.wielder is not None:
            bonus += self.ability_mod
        if bonus != 0:
            dam_str += '{:+d}'.format(bonus)
        return dam_str

    def __str__(self):
        return self.name

    def __repr__(self):
        return "\"{:s}\"".format(self.name)


class MeleeWeapon(Weapon):
    name = "Melee Weapons"
    ability = 'strength'


class RangedWeapon(Weapon):
    name = "Ranged Weapons"
    ability = 'dexterity'


class SimpleWeapon(Weapon):
    name = "Simple Weapons"


class MartialWeapon(Weapon):
    name = "Martial Weapons"
    

class Club(SimpleWeapon, MeleeWeapon):
    name = "Club"
    cost = "1 sp"
    base_damage = "1d4"
    damage_type = "b"
    weight = 2
    properties = "Light"
    ability = 'strength'


class Dagger(SimpleWeapon, MeleeWeapon):
    name = "Dagger"
    cost = "2 gp"
    base_damage = "1d4"
    damage_type = "p"
    weight = 1
    properties = "Finesse, light, thrown (range 20/60)"
    is_finesse = True
    ability = 'strength'


class Greatclub(SimpleWeapon, MeleeWeapon):
    name = "Greatclub"
    cost = "2 sp"
    base_damage = "1d8"
    damage_type = "b"
    weight = 10
    properties = "Two-handed"
    ability = 'strength'


class Handaxe(SimpleWeapon, MeleeWeapon):
    name = "Handaxe"
    cost = "5 gp"
    base_damage = "1d6"
    damage_type = "s"
    weight = 2
    properties = "Light, thrown (range 20/60)"
    ability = 'strength'


class Javelin(SimpleWeapon, MeleeWeapon):
    name = "Javelin"
    cost = "5 sp"
    base_damage = "1d6"
    damage_type = "p"
    weight = 2
    properties = "Thrown (range 30/120)"
    ability = 'strength'


class LightHammer(SimpleWeapon, MeleeWeapon):
    name = "Light hammer"
    cost = "2 gp"
    base_damage = "1d4"
    damage_type = "b"
    weight = 2
    properties = "Light, thrown (range 20/60)"
    ability = 'strength'


class Mace(SimpleWeapon, MeleeWeapon):
    name = "Mace"
    cost = "5 gp"
    base_damage = "1d6"
    damage_type = "b"
    weight = 4
    properties = ""
    ability = 'strength'


class Quarterstaff(SimpleWeapon, MeleeWeapon):
    name = "Quarterstaff"
    cost = "2 sp"
    base_damage = "1d6"
    damage_type = "b"
    weight = 4
    properties = "Versatile (1d8)"
    ability = 'strength'


class Sickle(SimpleWeapon, MeleeWeapon):
    name = "Sickle"
    cost = "1 gp"
    base_damage = "1d4"
    damage_type = "s"
    weight = 2
    properties = "Light"
    ability = 'strength'


class Spear(SimpleWeapon, MeleeWeapon):
    name = "Spear"
    cost = "1 gp"
    base_damage = "1d6"
    damage_type = "p"
    weight = 3
    properties = "Thrown (range 20/60), versatile (1d8)"
    ability = 'strength'


class LightCrossbow(SimpleWeapon, RangedWeapon):
    name = "Light crossbow"
    cost = "25 gp"
    base_damage = "1d8"
    damage_type = "p"
    weight = 5
    properties = "Ammunition (range 80/320), loading, two-handed"
    ability = 'dexterity'


class Dart(SimpleWeapon, RangedWeapon):
    name = "Dart"
    cost = "5 cp"
    base_damage = "1d4"
    damage_type = "p"
    weight = 0.25
    properties = "Finesse, thrown (range 20/60)"
    is_finesse = True
    ability = 'dexterity'


class Shortbow(SimpleWeapon, RangedWeapon):
    name = "Shortbow"
    cost = "25 gp"
    base_damage = "1d6"
    damage_type = "p"
    weight = 2
    properties = "Ammunition (range 80/320), two-handed"
    ability = 'dexterity'


class Sling(SimpleWeapon, RangedWeapon):
    name = "Sling"
    cost = "1 sp"
    base_damage = "1d4"
    damage_type = "b"
    weight = 0
    properties = "Ammunition (range 30/120)"
    ability = 'dexterity'


class Battleaxe(MartialWeapon, MeleeWeapon):
    name = "Battleaxe"
    cost = "10 gp"
    base_damage = "1d8"
    damage_type = "s"
    weight = 4
    properties = "Versatile (1d10)"
    ability = 'strength'


class Flail(MartialWeapon, MeleeWeapon):
    name = "Flail"
    cost = "10gp"
    base_damage = "1d8"
    damage_type = "b"
    weight = 2
    properties = ""
    ability = 'strength'


class Glaive(MartialWeapon, MeleeWeapon):
    name = "Glaive"
    cost = "20 gp"
    base_damage = "1d10"
    damage_type = "s"
    weight = 6
    properties = "Heavy, reach, two-handed"
    ability = 'strength'


class Greataxe(MartialWeapon, MeleeWeapon):
    name = "Greataxe"
    cost = "30 gp"
    base_damage = "1d12"
    damage_type = "s"
    weight = 7
    properties = "Heavy, two-handed"
    ability = 'strength'


class Greatsword(MartialWeapon, MeleeWeapon):
    name = "Greatsword"
    cost = "50 gp"
    base_damage = "2d6"
    damage_type = "s"
    weight = 6
    properties = "Heavy, two-handed"
    ability = 'strength'


class Halberd(MartialWeapon, MeleeWeapon):
    name = "Halberd"
    cost = "20 gp"
    base_damage = "1d10"
    damage_type = "s"
    weight = 6
    properties = "Heavy, reach, two-handed"
    ability = 'strength'


class Lance(MartialWeapon, MeleeWeapon):
    name = "Lance"
    cost = "10gp"
    base_damage = "1d12"
    damage_type = "p"
    weight = 6
    properties = "Reach, special"
    ability = 'strength'


class Longsword(MartialWeapon, MeleeWeapon):
    name = "Longsword"
    cost = "15 gp"
    base_damage = "1d8"
    damage_type = "s"
    weight = 3
    properties = "Versatile (1d10)"
    ability = 'strength'


class Maul(MartialWeapon, MeleeWeapon):
    name = "Maul"
    cost = "10 gp"
    base_damage = "2d6"
    damage_type = "b"
    weight = 10
    properties = "Heavy, two-handed"
    ability = 'strength'


class Morningstar(MartialWeapon, MeleeWeapon):
    name = "Morningstar"
    cost = "15 gp"
    base_damage = "1d8"
    damage_type = "p"
    weight = 4
    properties = ""
    ability = 'strength'


class Pike(MartialWeapon, MeleeWeapon):
    name = "Pike"
    cost = "5 gp"
    base_damage = "1d10"
    damage_type = "p"
    weight = 18
    properties = "Heavy, reach, two-handed"
    ability = 'strength'


class Rapier(MartialWeapon, MeleeWeapon):
    name = "Rapier"
    cost = "25 gp"
    base_damage = "1d8"
    damage_type = "p"
    weight = 2
    properties = "Finesse"
    is_finesse = True
    ability = 'strength'


class Scimitar(MartialWeapon, MeleeWeapon):
    name = "Scimitar"
    cost = "25 gp"
    base_damage = "1d6"
    damage_type = "s"
    weight = 3
    properties = "Finesse, light"
    is_finesse = True
    ability = 'strength'


class Shortsword(MartialWeapon, MeleeWeapon):
    name = "Shortsword"
    cost = "10 gp"
    base_damage = "1d6"
    damage_type = "p"
    weight = 0
    properties = "Finesse, light"
    is_finesse = True
    ability = 'strength'


class ThrowingHammer(MartialWeapon, MeleeWeapon):
    name = "Throwing Hammer"
    cost = "15 gp"
    base_damage = '1d6'
    damage_type = "b"
    weight = 4
    properties = "Thrown (range 60/120)"
    ability = "strength"


class Trident(MartialWeapon, MeleeWeapon):
    name = "Trident"
    cost = "5 gp"
    base_damage = "1d6"
    damage_type = "p"
    weight = 4
    properties = "Thrown (range 20/60), versatile (1d8)"
    ability = 'strength'


class WarPick(MartialWeapon, MeleeWeapon):
    name = "War pick"
    cost = "5 gp"
    base_damage = "1d8"
    damage_type = "p"
    weight = 2
    properties = ""
    ability = 'strength'


class Warhammer(MartialWeapon, MeleeWeapon):
    name = "Warhammer"
    cost = "15 gp"
    base_damage = "1d8"
    damage_type = "b"
    weight = 2
    properties = "Versatile (1d10)"
    ability = 'strength'


class Whip(MartialWeapon, MeleeWeapon):
    name = "Whip"
    cost = "2 gp"
    base_damage = "1d4"
    damage_type = "s"
    weight = 3
    properties = "Finesse, reach"
    is_finesse = True
    ability = 'strength'


class Blowgun(MartialWeapon, RangedWeapon):
    name = "Blowgun"
    cost = "10 gp"
    base_damage = "1"
    damage_type = "p"
    weight = 1
    properties = "Ammunition (range 25/100), loading"
    ability = 'dexterity'


class HandCrossbow(MartialWeapon, RangedWeapon):
    name = "Crossbow, hand"
    cost = "75 gp"
    base_damage = "1d6"
    damage_type = "p"
    weight = 3
    properties = "Ammunition (range 30/120), light, loading"
    ability = 'dexterity'


class HeavyCrossbow(MartialWeapon, RangedWeapon):
    name = "Crossbow, heavy"
    cost = "50 gp"
    base_damage = "1d10"
    damage_type = "p"
    weight = 18
    properties = "Ammunition (range 100/400), heaving, loading, two-handed"
    ability = 'dexterity'


class Longbow(MartialWeapon, RangedWeapon):
    name = "Longbow"
    cost = "50 gp"
    base_damage = "1d8"
    damage_type = "p"
    weight = 2
    properties = "Ammunition (range 150/600), heavy, two-handed"
    ability = 'dexterity'


class Net(MartialWeapon, RangedWeapon):
    name = "Net"
    cost = "1 gp"
    base_damage = "-"
    damage_type = ""
    weight = 3
    properties = "Special, thrown (range 5/15)"
    ability = 'strength'

    
class Unarmed(MeleeWeapon):
    name = "Unarmed"
    cost = "0 gp"
    base_damage = "1"
    damage_type = "b"
    weight = 0
    properties = ""
    ability = "strength"


class SunBolt(RangedWeapon):
    name = "Sun Bolt"
    cost = "0 gp"
    base_damage = "1d4"
    damage_type = "r"
    weight = 0
    properties = "(range 20/60)"
    ability = 'dexterity'

    
# Custom weapons
class HeavyPunch(MeleeWeapon):
    base_damage = "1d4"
    name = "Heavy Punch"
    damage_type = 'b'
    damage_bonus = 10  # Heavy weapon master
    attack_bonus = -5  # Heavy weapon master


class Bite(MeleeWeapon):
    name = "Bite"
    base_damage = "1d4"
    damage_type = 'p'
    cost = "0 gp"
    weight = 0
    properties = ""
    ability = "strength"


class Talons(MeleeWeapon):
    name = 'Talons'
    base_damage = '1d4'
    damage_type = 's'
    cost = '0 gp'
    weight = 0
    properties = ''
    ability = 'strength'


class Claws(MeleeWeapon):
    name = 'Claws'
    base_damage = '1d4'
    damage_type = 's'
    cost = '0 gp'
    weight = 0
    properties = ''
    ability = 'strength'


class Firearm(RangedWeapon):
    name = 'Firearm'
    ability = 'dexterity'
    damage_type = 'p'
    
    
class Blunderbuss(Firearm):
    name = 'Blunderbuss'
    base_damage = '2d8'
    cost = '300 gp'
    weight = 10
    properties = "Ammunition (range 15/60), Reload 1, Misfire 2"


class Pistol(Firearm):
    name = 'Pistol'
    base_damage = '1d10'
    cost = '150 gp'
    weight = 3
    properties = "Ammunition (range 60/240), Reload 4, Misfire 1"

    
class Musket(Firearm):
    name = 'Musket'
    base_damage = '1d12'
    cost = '300'
    weight = 10
    properties = "Ammunition (range 120/480), Two-Handed, Reload 1, Misfire 2"


# Magic Items
class FlameTongue(Greatsword):
    name = "Flame Tongue +1"
    damage_bonus = 1
    attack_bonus = 1
    base_damage = "4d6"
    damage_type = 'f'


class SpearOfLightning(Spear):
    name = "Spear +1"
    damage_bonus = 1
    attack_bonus = 1


# Some lists of weapons for easy proficiency resolution
simple_melee_weapons = (Club, Dagger, Greatclub, Handaxe, Javelin,
                        LightHammer, Mace, Quarterstaff, Sickle, Spear)
simple_ranged_weapons = (LightCrossbow, Dart, Shortbow, Sling)
simple_weapons = simple_melee_weapons + simple_ranged_weapons

martial_melee_weapons = (Battleaxe, Flail, Glaive, Greataxe,
                         Greatsword, Halberd, Lance, Longsword, Maul,
                         Morningstar, Pike, Rapier, Scimitar,
                         Shortsword, ThrowingHammer, Trident, WarPick,
                         Warhammer, Whip)
martial_ranged_weapons = (Blowgun, HandCrossbow, HeavyCrossbow,
                          Longbow, Net)
martial_weapons = martial_melee_weapons + martial_ranged_weapons

monk_weapons = (Shortsword, Unarmed, Club, Dagger, Handaxe, Javelin,
                LightHammer, Mace, Quarterstaff, Sickle, Spear, SunBolt)

firearms = (Firearm, Blunderbuss, Pistol, Musket)



from .features import Feature
from .. import armor, spells


# Many Classes
class Darkvision(Feature):
    """Accustomed to life underground, you have superior vision in dark and dim
    conditions. You can see in dim light within 60 feet of you as if it were
    bright light, and in darkness as if it were dim light. You can’t discern
    color in darkness, only shades of gray.

    """
    name = "Darkvision (60')"
    source = "Race"


class SuperiorDarkvision(Feature):
    """Accustomed to life underground, you have superior vision in dark and dim
    conditions. You can see in dim light within 120 feet of you as if it were
    bright light, and in darkness as if it were dim light. You can’t discern
    color in darkness, only shades of gray.

    """
    name = "Darkvision (120')"
    source = "Race"


class PowerfulBuild(Feature):
    """
    You count as one size larger when determining your carrying capacity and the weight you can push, drag, or lift.
    """
    name = "Powerful Build"
    source = "Race"

    
class Amphibious(Feature):
    """
    You can breath air and water

    """
    name = "Amphibious"
    source = "Race"

    
# Dwarves
class DwarvenResilience(Feature):
    """You have advantage on saving throws against poison, and you have resistance
    against poison damage

    """
    name = "Dwarven Resilience"
    source = "Race (Dwarf)"


class Stonecunning(Feature):
    """Whenever you make an Intelligence (History) check related to the origin of
    stonework, you are considered proficient in the History skill and add
    double your proficiency bonus to the check, instead of your normal
    proficiency bonus. Languages.

    """
    name = "Stonecunning"
    source = "Race (Dwarf)"

    
class DwarvenToughness(Feature):
    """
    Your hit point maximum
    increases by 1, and it increases by 1 every time you gain a level.

    """
    name = "DwarvenToughness"
    source = "Race (Hill Dwarf)"
    needs_implementation = True

    
# Elves
class FeyAncestry(Feature):
    """You have advantage on saving throws against being charmed, and magic can’t
    put you to sleep.

    """
    name = "Fey Ancestry"
    source = "Race (Elf)"


class Trance(Feature):
    """Elves don’t need to sleep. Instead, they meditate deeply, remaining
    semiconscious, for 4 hours a day. (The Common word for such meditation is
    “trance.”) While meditating, you can dream after a fashion; such dreams are
    actually mental exercises that have become reflexive through years of
    practice. After resting in this way, you gain the same benefit that a human
    does from 8 hours of sleep.

    """
    name = "Trance"
    source = "Race (Elf)"


class ElfCantrip(Feature):
    """You know one cantrip of your choice from the wizard spell
    list. Intelligence is your spellcasting ability for it.

    """
    name = "Cantrip"
    source = "Race (High-Elf)"
    needs_implementation = True


class MaskOfTheWild(Feature):
    """You can attempt to hide even when you are only lightly obscured by foliage,
    heavy rain, falling snow, mist, and other natural phenomena.

    """
    name = "Mask of the Wild"
    source = "Race (Wood Elf)"


class SunlightSensitivity(Feature):
    """You have disadvantage on attack rolls and on Wisdom (Perception) checks
    that rely on sight when you, the target of your attack, or whatever you are
    trying to perceive is in direct sunlight.

    """
    name = "Sunlight Sensitivity"
    source = "Race (Dark Elf)"


class DrowMagic(Feature):
    """You know the dancing lights cantrip.  When you reach 3rd level, you can
    cast the faerie fire spell once per day. When you reach 5th level, you can
    also cast the darkness spell once per day. Charisma is your spellcasting
    ability for these spells.

    """
    name = "Drow Magic"
    source = "Race (Dark Elf)"
    spells_known = spells_prepared = (spells.DancingLights,)


# Halflings
class Lucky(Feature):
    """When you roll a 1 on an attack roll, ability check, or saving throw, you can
    reroll the die and must use the new roll.

    """
    name = "Lucky"
    source = "Race (Halfling)"


class Brave(Feature):
    """You have advantage on saving throws against being frightened.

    """
    name = "Brave"
    source = "Race (Halfling)"


class HalflingNimbleness(Feature):
    """
    You can move through the space of any creature that is of a size larger than yours.
    """
    name = "Halfling Nimbleness"
    source = "Race (Halfling)"

    
class NaturallyStealthy(Feature):
    """You can attempt to hide even when you are obscured only by a creature that
    is at least one size larger than you.

    """
    name = "Naturally Stealthy"
    source = "Race (Lightfoot Halfling)"


class StoutResilience(Feature):
    """You have advantage on saving throws against poison, and you have resistance
    against poison damage.

    """
    name = "Stout Resilience"
    source = "Race (Stout Halfling)"


# Humans

# Dragonborn
class DraconicAncestry(Feature):
    """You have draconic ancestry. Choose one type of dragon from the Draconic
    Ancestry table. Your breath weapon and damage resistance are determined by the
    dragon type.

    Dragon -- Damage Type -- Breath Weapon

    Black -- Acid -- 5 by 30 ft. line (DEX save)

    Blue -- Lightning -- 5 by 30 ft. line (DEX save)

    Brass -- Fire -- 5 by 30 ft. line (DEX save)

    Bronze -- Lightning -- 5 by 30 ft. line (DEX save)

    Copper -- Acid -- 5 by 30 ft. line (DEX save)

    Gold -- Fire -- 15 ft. cone (DEX save)

    Green -- Poison -- 15 ft. cone (CON save)

    Red -- Fire -- 15 ft. cone (DEX save)

    Silver -- Cold -- 15 ft. cone (CON save)

    White -- White -- 15 ft. cone (CON save)

    """
    name = "Draconic Ancestry"
    source = "Race (Dragonborn)"


class BreathWeapon(Feature):
    """You can use your action to exhale destructive energy. Your draconic ancestry
    determines the size, shape, and damage type of the exhalation. When you use
    your breath weapon, each creature in the area of the exhalation must make a
    saving throw, the type of which is determined by your draconic
    ancestry. The DC for this saving throw equals 8 + your Constitution
    modifier + your proficiency bonus. A creature takes 2d6 damage on a failed
    save, and half as much damage on a successful one. The damage increases to
    3d6 at 6th level, 4d6 at 11th level, and 5d6 at 16th level. After you use
    your breath weapon, you can’t use it again until you complete a short or
    long rest. Damage

    """
    name = "Breath Weapon"
    source = "Race (Dragonborn)"


class DraconicResistance(Feature):
    """You have resistance to the damage type associated with your draconic
    ancestry. Languages.

    """
    name = "Damage Resistance"
    source = "Race (Dragonborn)"


# Gnomes
class GnomeCunning(Feature):
    """You have advantage on all Intelligence, Wisdom, and Charisma saving throws
    against magic.

    """
    name = "Gnome Cunning"
    source = "Race (Gnome)"


class NaturalIllusionist(Feature):
    """You know the minor illusion cantrip. Intelligence is your spellcasting
    ability for it.

    """
    name = "Natural Illusionist"
    source = "Race (Forest Gnome)"


class SpeakWithSmallBeasts(Feature):
    """Through sounds and gestures, you can communicate simple ideas with Small or
    smaller beasts. Forest gnomes love animals and often keep squirrels,
    badgers, rabbits, moles, woodpeckers, and other creatures as beloved pets.


    """
    name = "Speak with Small Beasts"
    source = "Race (Forest Gnome)"


class ArtificersLore(Feature):
    """Whenever you make an Intelligence (History) check related to magic items,
    alchemical objects, or technological devices, you can add twice your
    proficiency bonus, instead of any proficiency bonus you normally
    apply. Tinker.

    """
    name = "Artificer's Lore"
    source = "Race (Rock Gnome)"

    
class Tinker(Feature):
    """You have proficiency with artisan’s tools (tinker’s tools). Using those
    tools, you can spend 1 hour and 10 gp worth of materials to construct a
    Tiny clockwork device (AC 5, 1 hp). The device ceases to function after 24
    hours (unless you spend 1 hour repairing it to keep the device
    functioning), or when you use your action to dismantle it; at that time,
    you can reclaim the materials used to create it. You can have up to three
    such devices active at a time. When you create a device, choose one of the
    following options:

    *Clockwork Toy*: This toy is a clockwork animal, monster, or person, such as
    a frog, mouse, bird, dragon, or soldier. When placed on the ground, the toy
    moves 5 feet across the ground on each of your turns in a random
    direction. It makes noises as appropriate to the creature it represents.

    *Fire Starter*: The device produces a miniature flame, which you can use to
    light a candle, torch, or campfire. Using the device requires your action.

    *Music Box*: When opened, this music box plays a single song at a moderate
    volume. The box stops playing when it reaches the song’s end or when

    """


class StoneCamouflage(Feature):
    """
    You have advantage on Dexterity (stealth) checks to hide in rocky terrain.
    """
    name = "Stone Camouflage"
    source = "Race (Deep Gnome)"


# Half-Elves
    
# Half-Orcs
class RelentlessEndurance(Feature):
    """When you are reduced to 0 hit points but not killed outright, you can drop
    to 1 hit point instead. You can’t use this feature again until you finish a
    long rest.

    """
    name = "Relentless Endurance"
    source = "Race (Half-Orc)"


class SavageAttacks(Feature):
    """When you score a critical hit with a melee weapon attack, you can roll one
    of the weapon’s damage dice one additional time and add it to the extra
    damage of the critical hit.

    """
    name = "Savage Attacks"
    source = "Race (Half-Orc)"


# Tiefling
class HellishResistance(Feature):
    """You have resistance to fire damage.

    """
    name = "Hellish Resistance"
    source = "Race (Tiefling)"


class InfernalLegacy(Feature):
    """You know the thaumaturgy cantrip.  Once you reach 3rd level, you can cast
    the hellish rebuke spell once per day as a 2nd-level spell. Once you reach
    5th level, you can also cast the darkness spell once per day. Charisma is
    your spellcasting ability for these spells.
    
    """
    name = "Infernal Legacy"
    source = "Race (Tiefling)"
    spells_known = spells_prepared = (spells.Thaumaturgy,)


# Aasimar
class CelestialResistance(Feature):
    """
    You have resistance to necrotic damage and radiant damage.

    """
    name = "Celestial Resistance"
    source = "Race (Aasimar)"


class HealingHands(Feature):
    """As an action, you can touch a creature and cause it to regain a number of
    hit points equal to your level. Once you use this trait, you can't use it
    again until you finish a long rest.

    """
    name = "Healing Hands"
    source = "Race (Aasimar)"


class LightBearer(Feature):
    """You know the light cantrip. Charisma is your spellcasting ability for it.

    """
    name = "Light Bearer"
    source = "Race (Aasimar)"


class RadiantSoul(Feature):
    """Starting at 3rd level, you can use your action to unleash the divine energy
    within yourself, causing your eyes to glimmer and two luminous, incor-
    poreal wings to sprout from your back.

    Your transformation lasts for 1 minute or until you end it as a bonus
    action. During it, you have a flying speed of 30 feet, and once on each of
    your turns, you can deal extra radiant damage to one target when you deal
    damage to it with an attack or a spell. The extra radiant damage equals
    your level.

    Once you use this trait, you can't use it again until you finish a long
    rest.

    """
    name = "Radiant Soul"
    source = "Race (Protector Aasimar)"


class RadiantConsumption(Feature):
    """Starting at 3rd level, you can use your action to unleash the divine energy
    within yourself, causing a searing light to radiate from you, pour out of
    your eyes and mouth, and threaten to char you.

    Your transformation lasts for 1 minute or until you end ii as a bonus
    action. During it, you shed bright light in a 10-foot radius and dim light
    for an additional 10 feet,and at the end of each of your turns, you and
    each creature within 10 feet of you take radiant damage equal to half
    your level (rounded up). In addition, once on each of your turns, you can
    deal extra radiant damage to one target when you deal damage to it with an
    attack or a spell. The extra radiant damage equals your level.

    Once you use this trait, you can't use it again until you finish a long
    rest.
    
    """
    name = "Radiant Consumption"
    source = "Race (Scourge Aasimar)"


class NecroticShroud(Feature):
    """Starting at 3rd level, you can use your action to unleash the divine energy
    within yourself, causing your eyes to turn into pools of darkness and two
    skeletal, ghostly, flightless wings to sprout from your back. The instant
    you transform, other creatures within 10 feet of you that can see you must
    each succeed on a Charisma saving throw (DC 8 + your proficiency bonus +
    your Charisma modifier) or become frightened of you until the end of your
    next turn.

    Your transformation lasts for 1 minute or until you end it as a bonus
    action. During it, once on each of your turns, you can deal extra ne-
    crotic damage to one target when you deal damage to it with an attack or a
    spell. The extra necrotic damage equals your level.

    Once you use this trait, you can't use it again until you finish a long
    rest.

    """
    name = "Necrotic Shroud"
    source = "Race (Fallen Aasimar)"


# Firbolg
class FirbolgMagic(Feature):
    """You can cast detect magic and disguise self with this trait, using Wisdom
    as your spellcasting ability for them. Once you cast either spell, you
    can't cast it again with this trait until you finish a short or long
    rest. When you use this version of disguise self, you can seem up to 3 feet
    shorter than normal, allowing you to more easily blend in with humans and
    elves.

    """
    name = "Firbolg Magic"
    source = "Race (Firbolg)"


class HiddenStep(Feature):
    """As a bonus action, you can magically turn invisible until the start of your
    next turn or until you attack, make a damage roll, or force someone to make
    a saving throw. Once you use this trait, you can't use it again until you
    finish a short or long rest.

    """
    name = "Hidden Step"
    source = "Race (Firbolg)"


class SpeechOfBeastAndLeaf(Feature):
    """You have the ability to communicate in a limited manner with beasts and
    plants. They can understand the meaning of your words, though you have no
    special ability to understand them in return. You have advantage on all
    Charisma checks you make to influence them.

    """
    name = "Speech of Beast and Leaf"
    source = "Race (Firbolg)"


# Goliath
class StonesEndurance(Feature):
    """You can focus yourself to occasionally shrug off injury. When you take
    damage, you can use your reaction to roll a dl2. Add your Constitution
    modifier to the number rolled, and reduce the damage by that total. After
    you use this trait, you can't use it again until you finish a short or long
    rest.

    """
    name = "Stones Endurance"
    source = "Race (Goliath)"


class MountainBorn(Feature):
    """You're acclimated to high altitude, including elevations above 20,000
    feet. You're also naturally adapted to cold climates, as described in
    chapter 5 of the Dungeon Master's Guide.

    """
    name = "Mountain Born"
    source = "Race (Goliath)"


# Kenku
class ExpertForgery(Feature):
    """You can duplicate other creatures' handwriting and craftwork. You have
    advantage on all checks made to produce forgeries or duplicates of existing
    objects.

    """
    name = "Expert Forgery"
    source = "Race (Kenku)"

    
class Mimicry(Feature):
    """You can mimic sounds you have heard, including voices. A creature that
    hears the sounds you make can tell they are imitations with a successful
    Wisdom (Insight) check opposed by your Charisma (Deception) check.

    """
    name = "Mimicry"
    source = "Race (Kenku)"

    
# Lizardfolk
class CunningArtisan(Feature):
    """As part of a short rest, you can harvest bone and hide from a slain
    beast, construct, dragon, monstrosity, or plant creature of size Small or
    larger to create one of the following items: a shield, a club, a javelin,
    or ld4 darts or blowgun needles. To use this trait, you need a blade, such
    as a dagger, or appropriate artisan's tools, such as leatherworker's
    tools.

    """
    name = "Cunning Artisan"
    source = "Race (Lizardfolk)"


class HoldBreath(Feature):
    """
    You can hold your breath for up to 15 minutes at a time.
    """
    name = "Hold Breath"
    source = "Race (Lizardfolk)"


class NaturalArmor(Feature):
    """You have tough, scaly skin. When you aren't wearing armor, your AC is 13 +
    your Dexterity modifier. You can use your natural armor to determine your
    AC if the armor you wear would leave you with a lower AC. A shield's
    benefits apply as normal while you use your natural armor.

    """
    name = "Natural Armor"
    source = "Race (Lizardfolk)"

    def AC_func(self, char, **kwargs):
        """
        Implement the Natural Armor AC option
        """
        ac = 13 + char.dexterity.modifier
        if ((char.shield is not None)):
            ac += char.shield.base_armor_class
        return ac


class HungryJaws(Feature):
    """In battle, you can throw yourself into a vicious feeding frenzy. As a bonus
    action, you can make a special attack with your bite. If the attack hits,
    it deals its normal damage, and you gain temporary hit points (minimum of
    1) equal to your Constitution modifier, and you can't use this trait again
    until you finish a short or long rest.

    """
    name = "Hungry Jaws"
    source = "Race (Lizardfolk)"


# Tabaxi
class FelineAgility(Feature):
    """Your reflexes and agility allow you to meve with a burst of speed. When you
    move on your turn in combat, you can double your speed until the end of the
    turn. Once you use this trait, you can't use it again until you move O feet
    on one of your turns.

    """
    name = "Feline Agility"
    source = "Race (Tabaxi)"


# Triton

class ControlAirAndWater(Feature):
    """A child of the sea, you can call on the magic of elemental air and
    water. You can cast fog cloud with this trait. Starting at 3rd level, you
    can cast gust of wind with it, and starting at 5th level, you can also cast
    wall of water with it (see the spell in the sidebar). Once you cast a spell
    with this trait, you can't do so again until you finish a long
    rest. Charisma is your spellcasting ability for these spells.

    """
    name = "Control Air and Water"
    source = "Race (Triton)"


class EmissaryOfTheSea(Feature):
    """Aquatic beasts have an extraordinary affinity with your people. You can
    communicate simple ideas with beasts that can breathe water. They can
    understand the meaning of your words, though you have no special ability to
    understand them in return.

    """
    name = "Emissary Of The Sea"
    source = "Race (Triton)"


class GuardiansOfTheDepths(Feature):
    """Adapted to even the most extreme ocean depths, you have resistance to cold
    damage, and you ignore any of the drawbacks caused by a deep, underwater
    environment.

    """
    name = "Guardians of the Depths"
    source = "Race (Triton)"


# Aarakocra

# Genasi
class UnendingBreath(Feature):
    """You can hold your breath indefinitely while you’re not incapacitated.

    """
    name = "Unending Breath"
    source = "Race (Air Genasi)"


class MingleWithTheWind(Feature):
    """You can cast the levitate spell once with this trait, requiring no material
    components, and you regain the ability to cast it this way when you finish
    a long rest. Constitution is your spellcasting ability for this spell.

    """
    name = "Mingle with the Wind"
    source = "Race (Air Genasi)"


class EarthWalk(Feature):
    """You can move across difficult terrain made of earth or stone without
    expending extra movement.

    """
    name = "Earth Walk"
    source = "Race (Earth Genasi)"


class MergeWithStone(Feature):
    """You can cast the pass without trace spell once with this trait, requiring
    no material components, and you regain the ability to cast it this way when
    you finish a long rest. Constitution is your spellcasting ability for this
    spell.

    """
    name = "Merge with Stone"
    source = "Race (Earth Genasi)"


class FireResistance(Feature):
    """
    You have resistance to fire damage.
    """
    name = "Fire Resistance"
    source = "Race (Fire Genasi)"


class ReachToTheBlaze(Feature):
    """You know the produce flame cantrip. Once you reach 3rd level, you can cast
    the burning hands spell once with this trait as a 1st-level spell, and you
    regain the ability to cast it this way when you finish a long
    rest. Constitution is your spellcasting ability for these spells.

    """
    name = "Reach to the Blaze"
    source = "Race (Fire Genasi)"


class AcidResistance(Feature):
    """
    You have resistance to acid damage.
    
    """
    name = "Acid Resistance"
    source = "Race (Water Genasi)"


class CallToTheWave(Feature):
    """You know the shape water cantrip (see chapter 2 EEPC). When you reach 3rd level,
    you can cast the create or destroy water spell as a 2nd-level spell once
    with this trait, and you regain the ability to cast it this way when you
    finish a long rest. Constitution is your spellcasting ability for these
    spells.

    """
    name = "Call to the Wave"
    source = "Race (Water Genasi)"



from .spells import Spell


class MaddeningDarkness(Spell):
    """Magical darkness spreads from a point you choose within range to fill a 
    60—foot—radius sphere until the spell ends. The darkness spreads around corners.
     A creature with darkvision can’t see through this darkness. Nonmagical light, 
    as well as light created by spells of 8th level or lower, can't illuminate the 
    area. Shrieks, gibbering, and mad laughter can be heard within the sphere. 
    Whenever a creature starts its turn in the sphere, it must make a Wisdom saving 
    throw, taking 8d8 psychic damage on a failed save, or half as much damage on a 
    successful one.
    """
    name = "Maddening Darkness"
    level = 8
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ('V', 'M')
    materials = """A drop of pitch mixed with a drop of mercury"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Warlock', 'Wizard')


class Maelstrom(Spell):
    """(paper or leaf in the shape of a funnel)
    A mass of 5-foot-deep water appears and
     swirls in a 30-foot radius centered on a point you can see within range. The 
    point must be on ground or in a body of water. Until the spell ends, that area 
    is difficult terrain, and any creature that starts its turn there must succeed 
    on a Strength saving throw or take 6d6 bludgeoning damage and be pulled 10 feet 
    toward the center.
    """
    name = "Maelstrom"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid',)


class MageArmor(Spell):
    """You touch a willing creature who isn’t wearing armor, and a protective magical 
    force surrounds it until the spell ends. The target’s base AC becomes 13 + its 
    Dexterity modifier. The spell ends it if the target dons armor or if you dismiss
     the spell as an action.
    """
    name = "Mage Armor"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A piece of cured leather"""
    duration = "8 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Sorcerer', 'Wizard')


class MageHand(Spell):
    """A spectral, floating hand appears at a point you choose within range.
    The hand 
    lasts for the duration or until you dismiss it as an action. The hand vanishes 
    if it is ever more than 30 feet away from you or if you cast this spell again.
    
    
    You can use your action to control the hand. You can use the hand to manipulate 
    an object, open an unlocked door or container, stow or retrieve an item from an 
    open container, or pour the contents out of a vial. You can move the hand up to 
    30 feet each time you use it.
    
    The hand can’t attack, activate magical items, or
     carry more than 10 pounds.
    """
    name = "Mage Hand"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class MagicCircle(Spell):
    """You create a 10-foot-radius, 20-foot-tall cylinder of magical energy centered on
     a point on the ground that you can see within range. Glowing runes appear 
    wherever the cylinder intersects with the floor or other surface.
    
    Choose one or
     more of the following types of creatures: celestials, elementals, fey, fiends, 
    or undead. The circle affects a creature of the chosen type in the following 
    ways:
    
    * The creature can’t willingly enter the cylinder by nonmagical means. If
     the creature tries to use teleportation or interplanar travel to do so, it must
     first succeed on a Charisma saving throw.
    * The creature has disadvantage on 
    attack rolls against targets within the cylinder.
    * Targets within the cylinder 
    can’t be charmed, frightened, or possessed by the creature.
    
    When you cast this 
    spell, you can elect to cause its magic to operate in the reverse direction, 
    preventing a creature of the specified type from leaving the cylinder and 
    protecting targets outside it.
    
    At Higher Levels: When you cast this spell using
     a spell slot of 4th level or higher, the duration increases by 1 hour for each 
    slot level above 3rd.
    """
    name = "Magic Circle"
    level = 3
    casting_time = "1 minute"
    casting_range = "10 feet"
    components = ('V', 'S', 'M')
    materials = """Holy water or powdered silver and iron worth at least 100 gp, which the spell consumes"""
    duration = "1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Paladin', 'Warlock', 'Wizard')


class MagicJar(Spell):
    """Your body falls into a catatonic state as your soul leaves it and enters the 
    container you used for the spell’s material component. While your soul inhabits 
    the container, you are aware of your surroundings as if you were in the 
    container’s space. You can’t move or use reactions. The only action you can take
     is to project your soul up to 100 feet out of the container, either returning 
    to your living body (and ending the spell) or attempting to possess a humanoids 
    body.
    
    You can attempt to possess any humanoid within 100 feet of you that you 
    can see (creatures warded by a protection from evil and good or magic circle 
    spells can’t be possessed). The target must make a Charisma saving throw. On a 
    failure, your soul moves into the target’s body, and the target’s soul becomes 
    trapped in the container. On a success, the target resists your efforts to 
    possess it, and you can’t attempt to possess it again for 24 hours.
    
    Once you 
    possess a creature’s body, you control it. Your game statistics are replaced by 
    the statistics of the creature though you retain your alignment and your 
    Intelligence, Wisom, and Charisma scores. You retain the benefit of your own 
    class feature. If the target has any class levels, you can’t use any of its 
    class features.
    
    Meanwhile, the possessed creature’s soul can perceive from the 
    container using its own senses, but it can’t move or take actions at all.
    
    While
     possessing a body, you can use your action to return from the host body to the 
    container if it is within 100 feet of you, returning the host creature’s soul to
     its body. If the host body dies while you’re in it, the creature dies, and you 
    must make a Charisma saving throw against your own spellcasting DC. On a 
    success, you return to the container if it is within 100 feet of you. Otherwise,
     you die.
    
    If the container is destroyed or the spell ends, your soul 
    immediately returns to your body. If your body is more than 100 feet away from 
    you, or if your body is dead when you attempt to return to it, you die. If 
    another creature’s soul is in the container when it is destroyed, the creature’s
     soul returns to its body if the body is alive and within 100 feet. Otherwise, 
    that creature dies.
    
    When the spell ends, the container is destroyed.
    """
    name = "Magic Jar"
    level = 6
    casting_time = "1 minute"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """A gem, crystal, reliquary, or some other ornamental container worth at least 500 gp"""
    duration = "Until dispelled"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Wizard',)


class MagicMissile(Spell):
    """You create three glowing darts of magical force. Each dart hits a creature of 
    your choice that you can see within range. A dart deals 1d4 + 1 force damage to 
    its target. The darts all strike simultaneously and you can direct them to hit 
    one creature or several.
    
    At Higher Levels: When you cast this spell using a 
    spell slot of 2nd level or higher, the spell creates one more dart for each slot
     level above 1st.
    """
    name = "Magic Missile"
    level = 1
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class MagicMouth(Spell):
    """You implant a message within an object in range, a message that is uttered when 
    a trigger condition is met.
    Choose an object that you can see and that isn’t 
    being worn or carried by another creature. Then speak the message, which must be
     25 words or less, though it can be delivered over as long as 10 minutes. 
    Finally, determine the circumstance that will trigger the spell to deliver your 
    message.
    
    When that circumstance occurs, a magical mouth appears on the object 
    and recites the message in your voice and at the same volume you spoke. If the 
    object you chose has a mouth or something that looks like a mouth (for example, 
    the mouth of a statue), the magical mouth appears there so that words appear to 
    come from the object’s mouth. When you cast this spell, you can have the spell 
    end after it delivers its message, or it can remain and repeats its message 
    whenever the trigger occurs.
    
    The triggering circumstance can be as general or 
    as detailed as you like, though it must be based on visual or audible conditions
     that occur within 30 feet of the object. For example, you could instruct the 
    mouth to speak when any creature moves within 30 feet of the object or when a 
    silver bell rings within 30 feet of it.
    """
    name = "Magic Mouth"
    level = 2
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A small bit of honeycomb and jade dust worth at least 10 gp, which the spell consumes"""
    duration = "Until dispelled"
    ritual = True
    magic_school = "Illusion"
    classes = ('Bard', 'Wizard')


class MagicStone(Spell):
    """You touch one to three pebbles and imbue them with magic. You or someone else 
    can make a ranged spell attack with one of the pebbles by throwing it or hurling
     it with a sling. If thrown, a pebble has a range of 60 feet. If someone else 
    attacks with a pebble, that attacker adds your spellcasting ability modifier, 
    not the attacker’s, to the attack roll. On a hit, the target takes bludgeoning 
    damage equal to 1d6 + your spellcasting ability modifier. Whether the attack 
    hits or misses, the spell then ends on the stone.
    If you cast this spell again, 
    the spell ends on any pebbles still affected by your previous casting.
    """
    name = "Magic Stone"
    level = 0
    casting_time = "1 bonus action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Warlock')


class MagicWeapon(Spell):
    """You touch a nonmagical weapon. Until the spell ends, that weapon becomes a magic
     weapon with a +1 bonus to attack rolls and damage rolls.
    
    At Higher Levels: 
    When you cast this spell using a spell slot of 4th level or higher, the bonus 
    increases to +2.
    When you use a spell slot of 6th level or higher, the bonus 
    increases to +3.
    """
    name = "Magic Weapon"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Paladin', 'Wizard')


class MajorImage(Spell):
    """You create the image of an object, a creature, or some other visible phenomenon 
    that is no larger than a 20-foot cube.
    The image appears at a spot that you can 
    see within range and lasts for the duration. It seems completely real, including
     sounds, smells, and temperature appropriate to the thing depicted. You can’t 
    create sufficient heat or cold to cause damage, a sound loud enough to deal 
    thunder damage or deafen a creature, or a smell that might sicken a creature 
    (like a troglodyte’s stench).
    
    As long as you are within range of the illusion, 
    you can use your action to cause the image to move to any other spot within 
    range. As the image changes location, you can alter its appearance so that its 
    movements appear natural for the image. For example, if you create an image of a
     creature and move it, you can alter the image so that it appears to be walking.
     Similarly, you can cause the illusion to make different sounds at different 
    times, even making it carry on a conversation, for example.
    
    Physical 
    interaction with the image reveals it to be an illusion, because things can pass
     through it. A creature that uses its action to examine the image can determine 
    that it is an illusion with a successful Intelligence (Investigation) check 
    against your spell save DC. If a creature discerns the illusion for what it is, 
    the creature can see through the image, and its other sensory qualities become 
    faint to the creature.
    
    At Higher Levels: When you cast this spell using a spell
     slot of 6th level or higher, the spell lasts until dispelled, without requiring
     your concentration.
    """
    name = "Major Image"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of fleece"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class MassCureWounds(Spell):
    """A wave of healing energy washes out from a point of your choice within range.
    
    Choose up to six creatures in a 30-foot-radius sphere centered on that point. 
    Each target regains hit points equal to 3d8 + your spellcasting ability 
    modifier. This spell has no effect on undead or constructs.
    
    At Higher Levels: 
    When you cast this spell using a spell slot of 6th level or higher, the healing 
    increases by 1d8 for each slot level above 5th.
    """
    name = "Mass Cure Wounds"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard', 'Cleric', 'Druid')


class MassHeal(Spell):
    """A flood of healing energy flows from you into injured creatures around you. You 
    restore up to 700 hit points, divided as you choose among any number of 
    creatures that you can see within range. Creatures healed by this spell are also
     cured of all diseases and any effect making them blinded or deafened. This 
    spell has no effect on undead or constructs.
    """
    name = "Mass Heal"
    level = 9
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class MassHealingWord(Spell):
    """As you call out words of restoration, up to six creatures of your choice that 
    you can see within range regain hit points equal to 1d4 + your spellcasting 
    ability modifier. This spell has no effect on undead or constructs.
    
    At Higher 
    Levels: When you cast this spell using a spell slot of 4th level or higher, the 
    healing increases by 1d4 for each slot level above 3rd.
    """
    name = "Mass Healing Word"
    level = 3
    casting_time = "1 bonus action"
    casting_range = "60 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class MassPolymorph(Spell):
    """You transform up to ten creatures of your choice that you can see within range. 
    An unwilling target must succeed on a Wisdom saving throw to resist the 
    transformation. An unwilling shapechanger automatically succeeds on the save.
    
    Each target assumes a beast form of your choice, and you can choose the same 
    form or different ones for each target. The new form can be any beast you have 
    seen whose challenge rating is equal to or less than the target’s (or half the 
    target’s level, if the target doesn’t have a challenge rating). The target’s 
    game statistics, including mental ability scores, are replaced by the statistics
     of the chosen beast, but the target retains its hit points, alignment, and 
    personality.
    Each target gains a number of temporary hit points equal to the hit
     points of its new form. These temporary hit points can’t be replaced by 
    temporary hit points from another source. A target reverts to its normal form 
    when it has no more temporary hit points or it dies. If the spell ends before 
    then, the creature loses all its temporary hit points and reverts to its normal 
    form.
    The creature is limited in the actions it can perform by the nature of its
     new form. It can’t speak, cast spells, or do anything else that requires hands 
    or speech. The target’s gear melds into the new form.
    The target can’t activate,
     use, wield, or otherwise benefit from any of its equipment.
    """
    name = "Mass Polymorph"
    level = 9
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V.S', 'M')
    materials = """A caterpillar cocoon"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Wizard', 'Bard')


class MassSuggestion(Spell):
    """You suggest a course of activity (limited to a sentence or two) and magically 
    influence up to twelve creatures of your choice that you can see within range 
    and that can hear and understand you.
    Creatures that can’t be charmed are immune
     to this effect. The suggestion must be worded in such a manner as to make the 
    course of action sound reasonable. Asking the creature to stab itself, throw 
    itself onto a spear, immolate itself, or do some other obviously harmful act 
    automatically negates the effect of the spell.
    
    Each target must make a Wisdom 
    saving throw. On a failed save, it pursues the course of action you described to
     the best of its ability. The suggested course of action can continue for the 
    entire duration. If the suggested activity can be completed in a shorter time, 
    the spell ends when the subject finishes what it was asked to do.
    
    You can also 
    specify conditions that will trigger a special activity during the duration. For
     example, you might suggest that a group of soldiers give all their money to the
     first beggar they meet. If the condition isn’t met before the spell ends, the 
    activity isn’t performed.
    
    If you or any of your companions damage a creature 
    affected by this spell, the spell ends for that creature.
    
    At Higher Levels: 
    When you cast this spell using a 7th-level spell slot, the duration is 10 days.
    
    When you use an 8th-level spell slot, the duration is 30 days.
    When you use a
     9th-level spell slot, the duration is a year and a day.
    """
    name = "Mass Suggestion"
    level = 6
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'M')
    materials = """A snake’s tongue and either a bit of honeycomb or a drop of sweet oil"""
    duration = "24 hours"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class MaximiliansEarthenGrasp(Spell):
    """You choose a 5-foot-square unoccupied space on the ground that you can see 
    within range. A Medium hand made from compacted soil rises there and reaches
    for
     one creature you can see within 5 feet of it. The target must make a Strength 
    saving throw. On a failed save, the target takes 2d6 bludgeoning damage and is 
    restrained for the spell’s duration.
    As an action, you can cause the hand to 
    crush the restrained target, who must make a Strength saving throw. It takes 2d6
     bludgeoning damage on a failed save, or half as much damage on a successful 
    one.
    To break out, the restrained target can make a Strength check against your 
    spell save DC. On a success, the target escapes and is no longer restrained by 
    the hand.
    As an action, you can cause the hand to reach for a different creature
     or to move to a different unoccupied space within range. The hand releases a 
    restrained target if you do either.
    """
    name = "Maximilians Earthen Grasp"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A miniature hand sculpted from clay"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Wizard')


class Maze(Spell):
    """You banish a creature that you can see within range into a labyrinthine 
    demiplane. The target remains there for the duration or until it escapes the 
    maze.
    
    The target can use its action to attempt to escape. When it does so, it 
    makes a DC 20 Intelligence check. If it succeeds, it escapes, and the spell ends
     (a minotaur or goristro demon automatically succeeds).
    
    When the spell ends, 
    the target reappears in the space it left or, if that space is occupied, in the 
    nearest unoccupied space.
    """
    name = "Maze"
    level = 8
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Wizard',)


class MeldIntoStone(Spell):
    """You step into a stone object or surface large enough to fully contain your body,
     melding yourself and all the equipment you carry with the stone for the 
    duration.
    Using your movement, you step into the stone at a point you can touch.
     Nothing of your presence remains visible or otherwise detectable by nonmagical 
    senses.
    
    While merged with the stone, you can’t see what occurs outside it, and 
    any Wisdom (Perception) checks you make to hear sounds outside it are made with 
    disadvantage. You remain aware of the passage of time and can cast spells on 
    yourself while merged in the stone. You can use your movement to leave the stone
     where you entered it, which ends the spell. You otherwise can’t move.
    
    Minor 
    physical damage to the stone doesn’t harm you, but its partial destruction or a 
    change in its shape (to the extent that you no longer fit within it) expels you 
    and deals 6d6 bludgeoning damage to you. The stone’s complete destruction (or 
    transmutation into a different substance) expels you and deals 50 bludgeoning 
    damage to you. If expelled, you fall prone in an unoccupied space closest to 
    where you first entered.
    """
    name = "Meld Into Stone"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "8 hours"
    ritual = True
    magic_school = "Transmutation"
    classes = ('Cleric', 'Druid')


class MelfsAcidArrow(Spell):
    """A shimmering green arrow streaks toward a target within range and bursts in a 
    spray of acid.
    Make a ranged spell attack against the target. On a hit, the 
    target takes 4d4 acid damage immediately and 2d4 acid damage at the end of its 
    next turn. On a miss, the arrow splashes the target with acid for half as much 
    of the initial damage and no damage at the end of its next turn.
    
    At Higher 
    Levels: When you cast this spell using a spell slot of 3rd level or higher, the 
    damage (both initial and later) increases by 1d4 for each slot level above 2nd.
    """
    name = "Melfs Acid Arrow"
    level = 2
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """Powdered rhubarb leaf and an adder’s stomach"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Wizard',)


class MelfsMinuteMeteors(Spell):
    """(niter, sulfur, and pine tar formed into a bead)
    You create six tiny meteors in 
    your space. They float in the air and orbit you for the spell’s duration. When 
    you cast the spell—and as a bonus action on each of your turns thereafter—you 
    can expend one or two of the meteors, sending them streaking toward a point or 
    points you choose within 120 feet of you. Once a meteor reaches its destination 
    or impacts against a solid surface, the meteor explodes. Each creature within 5 
    feet of the point where the meteor explodes must make a Dexterity saving throw. 
    A creature takes 2d6 fire damage on a failed save, or half as much damage on a 
    successful one.
    At Higher Levels. When you cast this spell using a spell slot of
     4th level or higher, the number of meteors created increases by two for each 
    slot level above 3rd.
    """
    name = "Melfs Minute Meteors"
    level = 3
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class Mending(Spell):
    """This spell repairs a single break or tear in an object you touch, such as broken
     chain link, two halves of a broken key, a torn cloack, or a leaking wineskin.
    
    As long as the break or tear is no larger than 1 foot in any dimension, you mend
     it, leaving no trace of the former damage.
    
    This spell can physically repair a 
    magic item or construct, but the spell can’t restore magic to such an object.
    """
    name = "Mending"
    level = 0
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Two lodestones"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Cleric', 'Druid', 'Sorcerer', 'Wizard')


class MentalPrison(Spell):
    """You attempt to bind a creature within an illusory cell that only it perceives. 
    One creature you can see within range must make an Intelligence saving throw. 
    The target succeeds automatically if it is immune to being charmed. On a 
    successful save, the target takes 5d10 psychic damage, and the spell ends. On a 
    failed save, the target takes 5d10 psychic damage, and you make the area 
    immediately around the target’s space appear dangerous to it in some way. You 
    might cause the target to perceive itself as being surrounded by fire, floating 
    razors, or hideous maws filled with dripping teeth. Whatever form the illusion 
    takes, the target can’t see or hear anything beyond it and is restrained for the
     spell’s duration. If the target is moved out of the illusion, makes a melee 
    attack through it, or reaches any part of its body through it, the target takes 
    10d10 psychic damage, and the spell ends.
    """
    name = "Mental Prison"
    level = 6
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('S',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class Message(Spell):
    """You point your finger toward a creature within range and whisper a message.
    The 
    target (and only the target) hears the message and can reply in a whisper that 
    only you can hear.
    
    You can cast this spell through solid objects if you are 
    familiar with the target and know it is beyond the barrier. Magical silence, 1 
    foot of stone, 1 inch of common metal, a thin sheet of lead, or 3 feet of wood 
    blocks the spell. The spell doesn’t have to follow a straight line and can 
    travel freely around corners or through openings.
    """
    name = "Message"
    level = 0
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A short piece of copper wire"""
    duration = "1 round"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class MeteorSwarm(Spell):
    """Blazing orbs of fire plummet to the ground at four different points you can see 
    within range.
    Each creature in a 40-foot-radius sphere centered on each point 
    you choose must make a Dexterity saving throw. The sphere spreads around 
    corners. A creature takes 20d6 fire damage and 20d6 bludgeoning damage on a 
    failed save, or half as much damage on a sucessful one. A creature in the area 
    of more than one fiery burst is affected only once.
    
    The spell damages objects 
    in the area and ignites flammable objects that aren’t being worn or carried.
    """
    name = "Meteor Swarm"
    level = 9
    casting_time = "1 action"
    casting_range = "1 mile"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class MightyFortress(Spell):
    """A fortress of stone erupts from a square area of ground of your choice that you 
    can see within range. The area is 120 feet on each side, and it must not have 
    any buildings or other structures on it. Any creatures in the area are 
    harmlessly lifted up as the fortress rises.
    The fortress has four turrets with 
    square bases, each one 20 feet on a side and 30 feet tall, with one turret on 
    each corner. The turrets are connected to each other by stone walls that are 
    each 80 feet long, creating an enclosed area. Each wall is 1 foot thick and is 
    composed of panels that are 10 feet wide and 20 feet tall. Each panel is 
    contiguous with two other panels or one other panel and a turret. You can place 
    up to four stone doors in the fortress’s outer wall.
    A small keep stands inside 
    the enclosed area. The keep has a square base that is 50 feet on each side, and 
    it has three floors with 10-foot-high ceilings. Each of the floors can be 
    divided into as many rooms as you like, provided each room is at least 5 feet on
     each side. The floors of the keep are connected by stone staircases, its walls 
    are 6 inches thick, and interior rooms can have stone doors or open archways as 
    you choose. The keep is furnished and decorated however you like, and it 
    contains sufficient food to serve a nine-course banquet for up to 100 people 
    each day. Furnishings, food, and other objects created by this spell crumble to 
    dust if removed from the fortress.
    A staff of one hundred invisible servants 
    obeys anycommand given to them by creatures you designate when you cast the 
    spell. Each servant functions as if created by the unseen servant spell.
    The 
    walls, turrets, and keep are all made of stone that can be damaged. Each 
    10—foot—bya10-foot section of stone has AC 15 and 30 hit points per inch of 
    thickness. It is immune to poison and psychic damage. Reducing a section of 
    stone to 0 hit points destroys it and might cause connected sections to buckle 
    and collapse at the DM’s discretion.
    After 7 days or when you cast this spell 
    somewhere else, the fortress harmlessly crumbles and sinks back into the ground,
     leaving any creatures that were inside it safely on the ground.
    Casting this 
    spell on the same spot once every 7 days for a year makes the fortress 
    permanent.
    """
    name = "Mighty Fortress"
    level = 8
    casting_time = "1 minute"
    casting_range = "1 mile"
    components = ('V', 'S', 'M')
    materials = """A diamond worth at least 500 gp, which the spell consumes"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Wizard',)


class MindBlank(Spell):
    """Until the spell ends, one willing creature you touch is immune to psychic 
    damage, any effect that would sense its emotions or read its thoughts, 
    divination spells, and the charmed condition. The spell even foils wish spells 
    and spells or effects of similar power used to affect the target’s mind or to 
    gain information about the target.
    """
    name = "Mind Blank"
    level = 8
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "24 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Bard', 'Wizard')


class MindSpike(Spell):
    """You reach into the mind of one creature you can see within range. The target 
    must make a Wisdom saving throw, taking 3d8 psychic damage on a failed save, or 
    half as much damage on a successful one. On a failed save, you also always know 
    the target's location until the spell ends, but only while the two of you are on
     the same plane of existence. While you have this knowledge, the target can’t 
    become hidden from you, and if it’s invisible, it gains no benefit from that 
    condition against you.
    
    At Higher Levels: When you cast this spell using a spell
     slot of 3rd level or higher, the damage increases by 1d8 for each slot level 
    above 2nd.
    """
    name = "Mind Spike"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('S',)
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Divination"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class MinorIllusion(Spell):
    """You create a sound or an image of an object within range that lasts for the 
    duration. The illusion also ends if you dismiss it as an action or cast this 
    spell again.
    
    If you create a sound, its volume can range from a whisper to a 
    scream. It can be your voice, someone else’s voice, a lion’s roar, a beating of 
    drums, or any other sound you choose. The sound continues unabated throughout 
    the duration, or you can make discrete sounds at different times before the 
    spell ends.
    
    If you create an image of an object such as a chair, muddy 
    footprints, or a small chest it must be no larger than a 5-foot cube. The image 
    can’t create sound, light, smell, or any other sensory effect. Physical 
    interaction with the image reveals it to be an illusion, because things can pass
     through it.
    
    If a creature uses its action to examine the sound or image, the 
    creature can determine that it is an illusion with a successful Intelligence 
    (Investigation) check against your spell save DC. If a creature discerns the 
    illusion for what it is, the illusion becomes faint to the creature.
    """
    name = "Minor Illusion"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('S', 'M')
    materials = """A bit of fleece"""
    duration = "1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class MirageArcane(Spell):
    """You make terrain in an area up to 1 mile square look, sound, smell, and even 
    feel like some other sort of terrain.
    The terrain’s general shape remains the 
    same, however. Open fields or a road could be made to resemble a swamp, hill, 
    crevasse, or some other difficult or impassable terrain. A pond can be made to 
    seem like a grassy meadow, a precipice like a gentle slope, or a rock-strewn 
    gully like a wide and smooth road.
    
    Similarly, you can alter the appearance of 
    structures, or add them where none are present. The spell doesn’t disguise, 
    conceal, or add creatures.
    
    The illusion includes audible, visual, tactile, and 
    olfactory elements, so it can turn clear ground into difficult terrain (or vice 
    versa) or otherwise impede movement through the area. Any piece of the illusory 
    terrain (such as a rock or stick) that is removed from the spell’s area 
    disappears immediately.
    
    Creatures with truesight can see through the illusion 
    to the terrain’s true form however, all other elements of the illusion remain, 
    so while the creature is aware of the illusion’s presence, the creature can 
    still physically interact with the illusion.
    """
    name = "Mirage Arcane"
    level = 7
    casting_time = "10 minutes"
    casting_range = "Sight"
    components = ('V', 'S')
    materials = """"""
    duration = "10 days"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Druid', 'Wizard')


class MirrorImage(Spell):
    """Three illusory duplicates of yourself appear in your space.
    Until the spell 
    ends, the duplicates move with you and mimic your actions, shifting position so 
    it’s impossible to track which image is real. You can use your action to dismiss
     the illusory duplicates.
    
    Each time a creature targets you with an attack 
    during the spell’s duration, roll a d20 to determine whether the attack instead 
    targets one of your duplicates.
    
    If you have three duplicates, you must roll a 6
     or higher to change the attack’s target to a duplicate. With two duplicates, 
    you must roll an 8 or higher. With one duplicate, you must roll an 11 or higher.
    
    
    A duplicate’s AC equals 10 + your Dexterity modifier. If an attack hits a 
    duplicate, the duplicate is destroyed. A duplicate can be destroyed only by an 
    attack that hits it. It ignores all other damage and effects. The spell ends 
    when all three duplicates are destroyed.
    
    A creature is unaffected by this spell
     if it can’t see, if it relies on senses other than sight, such as blindsight, 
    or if it can perceive illusions as false, as with truesight.
    """
    name = "Mirror Image"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class Mislead(Spell):
    """You become invisible at the same time that an illusory double of you appears 
    where you are standing. The double lasts for the duration, but the invisibility 
    ends if you attack or cast a spell.
    
    You can use your action to move your 
    illusory double up to twice your speed and make it gesture, speak, and behave in
     whatever way you choose.
    
    You can see through its eyes and hear through its 
    ears as if you were located where it is. On each of your turns as a bonus 
    action, you can switch from using its senses to using your own, or back again. 
    While you are using its senses, you are blinded and deafened in regard to your 
    own surroundings.
    """
    name = "Mislead"
    level = 5
    casting_time = "1 action"
    casting_range = "Self"
    components = ('S',)
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Wizard')


class MistyStep(Spell):
    """Briefly surrounded by silvery mist, you teleport up to 30 feet to an unoccupied 
    space that you can see.
    """
    name = "Misty Step"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class ModifyMemory(Spell):
    """You attempt to reshape another creature’s memories.
    One creature that you can 
    see must make a Wisdom saving throw. If you are fighting the creature, it has 
    advantage on the saving throw. On a failed save, the target becomes charmed by 
    you for the duration. The charmed target is incapacitated and unaware of its 
    surroundings, though it can still hear you. If it takes any damage or is 
    targeted by another spell, this spell ends, and none of the target’s memories 
    are modified.
    
    While this charm lasts, you can affect the target’s memory of an 
    event that it experienced within the last 24 hours and that lasted no more than 
    10 minutes. You can permanently eliminate all memory of the event, allow the 
    target to recall the event with perfect clarity and exacting detail, change its 
    memory of the details of the event, or create a memory of some other event.
    
    You
     must speak to the target to describe how its memories are affected, and it must
     be able to understand your language for the modified memories to take root. Its
     mind fills in any gaps in the details of your description. If the spell ends 
    before you have finished describing the modified memories, the creature’s memory
     isn’t altered. Otherwise, the modified memories take hold when the spell ends.
    
    
    A modified memory doesn’t necessarily affect how a creature behaves, 
    particularly if the memory contradicts the creature’s natural inclinations, 
    alignment, or beliefs. An illogical modified memory, such as implanting a memory
     of how much the creature enjoyed dousing itself in acid, is dismissed, perhaps 
    as a bad dream. The DM might deem a modified memory too nonsensical to affect a 
    creature in a significant manner.
    
    A remove curse or greater restoration spell 
    cast on the target restores the creature’s true memory.
    
    At Higher Levels: If 
    you cast this spell using a spell slot of 6th level or higher, you can alter the
     target’s memories of an event that took place up to 7 days ago (6th level), 30 
    days ago (7th level), 1 year ago (8th level), or any time in the creature’s past
     (9th level).
    """
    name = "Modify Memory"
    level = 5
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Wizard')


class MoldEarth(Spell):
    """You choose a portion of dirt or stone that you can see within range and that 
    fits within a 5-foot cube. You manipulate it in one of the following ways:
    • If 
    you target an area of loose earth, you can instantaneously excavate it, move it 
    along the ground, and deposit it up to 5 feet away. This movement doesn’t have 
    enough force to cause damage.
    • You cause shapes, colors, or both to appear on 
    the dirt or stone, spelling out words, creating images, or shaping patterns. The
     changes last for 1 hour.
    • If the dirt or stone you target is on the ground, 
    you cause it to become difficult terrain. Alternatively, you can cause the 
    ground to become normal terrain if it is already difficult terrain. This change 
    lasts for 1 hour. If you cast this spell multiple times, you can have no more 
    than two of its non-instantaneous effects active at a time, and you can dismiss 
    such an effect as an action.
    """
    name = "Mold Earth"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('S',)
    materials = """"""
    duration = "Instantaneous or 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class Moonbeam(Spell):
    """A silvery beam of pale light shines down in a 5-footradius, 40-foot-high 
    cylinder centered on a point within range. Until the spell ends, dim light fills
     the cylinder.
    
    
    When a creature enters the spell’s area for the first time on a
     turn or starts its turn there, it is engulfed in ghostly flames that cause 
    searing pain, and it must make a Constitution saving throw. It takes 2d10 
    radiant damage on a failed save, or half as much damage on a successful one.
    
    A 
    shapechanger makes its saving throw with disadvantage. If it fails, it also 
    instantly reverts to its original form and can’t assume a different form until 
    it leaves the spell’s light.
    
    On each of your turns after you cast this spell, 
    you can use an action to move the beam 60 feet in any direction.
    
    At Higher 
    Levels: When you cast this spell using aspell slot of 3rd level or higher, the 
    damage increases by 1d1O for each slot level above 2nd.
    """
    name = "Moonbeam"
    level = 2
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """Several seeds of any moonseed plant and a piece of opalescent feldspar"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid',)


class MordenkainensFaithfulHound(Spell):
    """You conjure a phantom watchdog in an unoccupied space that you can see within 
    range, where it remains for the duration, until you dismiss it as an action, or 
    until you move more than 100 feet away from it.
    
    The hound is invisible to all 
    creatures except you and can’t be harmed. When a Small or larger creature comes 
    within 30 feet of it without first speaking the password that you specify when 
    you cast this spell, the hound starts barking loudly. The hound sees invisible 
    creatures and can see into the Ethereal Plane. It ignores illusions.
    
    At the 
    start of each of your turns, the hound attempts to bite one creature within 5 
    feet of it that is hostile to you. The hound’s attack bonus is equal to your 
    spellcasting ability modifier + your proficiency bonus. On a hit, it deals 4d8 
    piercing damage.
    """
    name = "Mordenkainens Faithful Hound"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A tiny silver whistle, a piece of bone, and a thread"""
    duration = "8 hours"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Wizard',)


class MordenkainensMagnificentMansion(Spell):
    """You conjure an extradimensional dwelling in range that lasts for the duration.
    
    You choose where its one entrance is located. The entrance shimmers faintly and 
    is 5 feet w ide and 10 feet tall. You and any creature you designate when you 
    cast the spell can enter the extradimensional dwelling as long as the portal 
    remains open. You can open or close the portal if you are within 30 feet of it. 
    While closed, the portal is invisible.
    
    Beyond the portal is a magnificent foyer
     with numerous chambers beyond. The atmosphere is clean, fresh, and warm.
    
    You 
    can create any floor plan you like, but the space can’t exceed 50 cubes, each 
    cube being 10 feet on each side. The place is furnished and decorated as you 
    choose. It contains sufficient food to serve a ninecourse banquet for up to 100 
    people. A staff of 100 near-transparent servants attends all who enter. You 
    decide the visual appearance of these servants and their attire. They are 
    completely obedient to your orders. Each servant can perform any task a normal 
    human servant could perform, but they can’t attack or take any action that would
     directly harm another creature. Thus the servants can fetch things, clean, 
    mend, fold clothes, light fires, serve food, pour wine, and so on. The servants 
    can go anywhere in the mansion but can’t leave it. Furnishings and other objects
     created by this spell dissipate into smoke if removed from the mansion. When 
    the spell ends, any creatures inside the extradimensional space are expelled 
    into the open spaces nearest to the entrance.
    """
    name = "Mordenkainens Magnificent Mansion"
    level = 7
    casting_time = "1 minute"
    casting_range = "300 feet"
    components = ('V', 'S', 'M')
    materials = """A miniature portal carved from ivory, a small piece of polished marble, and a tiny silver spoon, each item worth at least 5 gp"""
    duration = "24 hours"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Bard', 'Wizard')


class MordenkainensPrivateSanctum(Spell):
    """You make an area within range magically secure.
    The area is a cube that can be 
    as small as 5 feet to as large as 100 feet on each side. The spell lasts for the
     duration or until you use an action to dismiss it.
    
    When you cast the spell, 
    you decide what sort of security the spell provides, choosing any or all of the 
    following properties:
    * Sound can’t pass through the barrier at the edge of the 
    warded area.
    * The barrier of the warded area appears dark and foggy, preventing
     vision (including darkvision) through it.
    * Sensors created by divination 
    spells can’t appear inside the protected area or pass through the barrier at its
     perimeter.
    * Creatures in the area can’t be targeted by divination spells.
    * 
    Nothing can teleport into or out of the warded area.
    * Planar travel is blocked 
    within the warded area.
    
    Casting this spell on the same spot every day for a 
    year makes this effect permanent.
    
    At Higher Levels: When you cast this spell 
    using a spell slot of 5th level or higher, you can increase the size of the cube
     by 100 feet for each slot level beyond 4th. Thus you could protect a cube that 
    can be up to 200 feet on one side by using a spell slot o f 5th level.
    """
    name = "Mordenkainens Private Sanctum"
    level = 4
    casting_time = "10 minutes"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A thin sheet of lead, a piece of opaque glass, a wad of cotton or cloth, and powdered chrysolite"""
    duration = "24 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Wizard',)


class MordenkainensSword(Spell):
    """You create a sword-shaped plane of force that hovers within range. It lasts for 
    the duration.
    
    When the sword appears, you make a melee spell attack against a 
    target of your choice within 5 feet of the sword. On a hit. the target takes 
    3d10 force damage. Until the spell ends, you can use a bonus action on each of 
    your turns to move the sword up to 20 feet to a spot you can see and repeat this
     attack against the same target or a different one.
    """
    name = "Mordenkainens Sword"
    level = 7
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A miniature platinum sword with a grip and pommel of copper and zinc, worth 250 gp"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard', 'Wizard')


class MoveEarth(Spell):
    """Choose an area of terrain no larger than 40 feet on a side within range. You can
     reshape dirt, sand, or clay in the area in any manner you choose for the 
    duration. You can raise or lower the area’s elevation, create or fill in a 
    trench, erect or flatten a wall, or form a pillar. The extent of any such 
    changes can’t exceed half the area’s largest dimension. So, if you affect a 
    40-foot square, you can create a pillar up to 20 feet high, raise or lower the 
    square’s elevation by up to 20 feet, dig a trench up to 20 feet deep, and so on.
     It takes 10 minutes for these changes to complete.
    
    At the end of every 10 
    minutes you spend concentrating on the spell, you can choose a new area of 
    terrain to affect.
    
    Because the terrain’s transformation occurs slowly, 
    creatures in the area can’t usually be trapped or injured by the ground’s 
    movement.
    
    This spell can’t manipulate natural stone or stone construction. 
    Rocks and structures shift to accommodate the new terrain. If the way you shape 
    the terrain would make a structure unstable, it might collapse.
    
    Similarly, this
     spell doesn’t directly affect plant growth. The moved earth carries any plants 
    along with it.
    """
    name = "Move Earth"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """An iron blade and a small bag containing a mixture of soils—clay, loam, and sand"""
    duration = "Concentration, up to 2 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Sorcerer', 'Wizard')





from .spells import Spell


class ZephyrStrike(Spell):
    """You move like the wind. Until the spell ends, your movement doesn’t provoke 
    opportunity attacks.
    Once before the spell ends, you can give yourself advantage
     on one weapon attack roll on your turn. That attack deals an extra 1d8 force 
    damage on a hit. Whether you hit or miss, your walking speed increases by 30 
    feet until the end of that turn.
    """
    name = "Zephyr Strike"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Ranger',)


class ZoneOfTruth(Spell):
    """You create a magical zone that guards against deception in a 15-foot-radius 
    sphere centered on a point of your choice within range.
    Until the spell ends, a 
    creature that enters the spell’s area for the first time on a turn or starts its
     turn there must make a Charisma saving throw. On a failed save, a creature 
    can’t speak a deliberate lie while in the radius. You know whether each creature
     succeeds or fails on its saving throw.
    
    An affected creature is aware of the 
    spell and can thus avoid answering questions to which it would normally respond 
    with a lie. Such creatures can be evasive in its answers as long as it remains 
    within the boundaries of the truth.
    """
    name = "Zone Of Truth"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "10 minutes"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Cleric', 'Paladin')





from .. import (features, weapons)
from .classes import (CharClass, SubClass)
from collections import defaultdict


# PHB
class BerserkerPath(SubClass):
    """For some barbarians, rage is a means to an end-—that end being
    violence. The Path of the Berserker is a path of untrammeled fury, slick
    with blood. As you enter the berserker’s rage, you thrill in the chaos of
    battle, heedless of your own health or well-being.

    """
    name = "Path of the Berserker"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.Frenzy]
    features_by_level[6] = [features.MindlessRage]
    features_by_level[10] = [features.IntimidatingPresence]
    features_by_level[14] = [features.Retaliation]


class TotemWarriorPath(SubClass):
    """The Path of the Totem Warrior is a spiritual journey, as the barbarian
    accepts a spirit animal as guide, protector, and inspiration. In battle,
    your totem spirit fills you with supernatural might, adding magical fuel to
    your barbarian rage.

    Most barbarian tribes consider a totem animal to be kin to a particular
    clan. In such cases, it is unusual for an individual to have more than one
    totem animal spirit, though exceptions exist.

    """
    name = "Path of the Totem Warrior"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.SpiritSeeker, features.TotemSpirit]
    features_by_level[6] = [features.BeastAspect]
    features_by_level[10] = [features.SpiritWalker]
    features_by_level[14] = [features.TotemicAttunement]


# SCAG
class BattleragerPath(SubClass):
    """Known as Kuldjargh (literally "axe idiot") in Dwarvish, battleragers are
    dwarf followers of the gods of war and take the Path of the
    Battlerager. They specialize in wearing bulky, spiked armor and throwing
    themselves into combat, striking with their body itself and giving
    themselves over to the fury of battle.

    """
    name = "Path of the Battlerager"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.BattleragerArmor]
    features_by_level[6] = [features.RecklessAbandon]
    features_by_level[10] = [features.BattleragerCharge]
    features_by_level[14] = [features.SpikedRetribution]


# XGTE
class AncestralGuardianPath(SubClass):
    """Some barbarians hail from cultures that revere their ancestors. These
    tribes teach that the warriors of the past linger in the world as mighty
    spirits, who can guide and protect the living. When a barbarian who follows
    this path rages, the barbarian contacts the spirit world and calls on these
    guardian spirits for aid.

    Barbarians who draw on their ancestral guardians can better fight to
    protect their tribes and their allies. In order to cement ties to their
    ancestral guardians, barbarians who follow this path cover themselves in
    elabo— rate tattoos that celebrate their ancestors’ deeds. These tattoos
    tell sagas of victories against terrible monsters and other fearsome
    rivals.

    """
    name = "Path of the Ancestral Guardian"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.AncestralProtectors]
    features_by_level[6] = [features.SpiritShield]
    features_by_level[10] = [features.ConsultTheSpirits]
    features_by_level[14] = [features.VengefulAncestors]

    
class StormHeraldPath(SubClass):
    """All barbarians harbor a fury within. Their rage grants them superior
    strength, durability, and speed. Barbarians who follow the Path of the
    Storm Herald learn to transform that rage into a mantle of primal magic,
    which swirls around them. When in a fury, a barbarian ofthis path taps into
    the forces of nature to create powerful magical effects.

    Storm heralds are typically elite champions who train alongside druids,
    rangers, and others sworn to protect nature. Other storm heralds hone their
    craft in lodges in regions wracked by storms, in the frozen reaches at the
    world’s end, or deep in the hottest deserts.

    """
    name = "Path of the Storm Herald"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.StormAura]
    features_by_level[6] = [features.StormSoul]
    features_by_level[10] = [features.ShieldingStorm]
    features_by_level[14] = [features.RagingStorm]


class ZealotPath(SubClass):
    """Some deities inspire their followers to pitch themselves into a ferocious
    battle fury. These barbarians are zealots—warriors who channel their rage
    into powerful disn plays of divine power.

    A variety of gods across the worlds of D&D inspire their followers to
    embrace this path. Tempus from the Forgotten Realms and Hextor and Erythnul
    of Greyhawk are all prime examples. In general, the gods who inspire
    zealots are deities of combat, destruction, and violence. Not all are evil,
    but few are good

    """
    name = "Path of the Zealot"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.DivineFury, features.WarriorOfTheGods]
    features_by_level[6] = [features.FanaticalFocus]
    features_by_level[10] = [features.ZealousPresence]
    features_by_level[14] = [features.RageBeyondDeath]
    

class Barbarian(CharClass):
    name = 'Barbarian'
    hit_dice_faces = 12
    subclass_select_level = 3
    saving_throw_proficiencies = ('strength', 'constitution')
    primary_abilities = ('strength',)
    weapon_proficiencies = (weapons.SimpleWeapon, weapons.MartialWeapon)
    _proficiencies_text = ('light armor', 'medium armor', 'shields',
                           'simple weapons', 'martial weapons')
    multiclass_weapon_proficiencies = weapon_proficiencies
    _multiclass_proficiencies_text = ('shields', 'simple weapons',
                                      'martial weapons')
    class_skill_choices = ('Animal Handling', 'Athletics',
                           'Intimidation', 'Nature', 'Perception', 'Survival')
    subclasses_available = (BerserkerPath, TotemWarriorPath, BattleragerPath,
                            AncestralGuardianPath, StormHeraldPath, ZealotPath)
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.Rage, features.UnarmoredDefenseBarbarian]
    features_by_level[2] = [features.RecklessAttack, features.DangerSense]
    features_by_level[5] = [features.ExtraAttackBarbarian,
                            features.FastMovement]
    features_by_level[7] = [features.FeralInstinct]
    features_by_level[9] = [features.BrutalCritical]
    features_by_level[11] = [features.RelentlessRage]
    features_by_level[15] = [features.PersistentRage]
    features_by_level[18] = [features.IndomitableMight]
    features_by_level[20] = [features.PrimalChampion]



import re
from collections import namedtuple

from .exceptions import DiceError

dice_re = re.compile('(\d+)d(\d+)', flags=re.I)
Dice = namedtuple('Dice', ('num', 'faces'))

def read_dice_str(dice_str):
    """Interpret a D&D dice string, eg. 3d10.
    
    Returns
    -------
    dice : tuple
      A named tuple with the scheme (num, faces), so '3d10' return
      (num=3, faces=10)
    
    """
    match = dice_re.match(dice_str)
    if match is None:
        raise DiceError(f"Cannot interpret dice string {dice_str}")
    dice = Dice(num=int(match.group(1)),
                faces=int(match.group(2)))
    return dice



from .spells import Spell


class Jump(Spell):
    """You touch a creature. The creature’s jump distance is tripled until the spell 
    ends.
    """
    name = "Jump"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A grasshopper’s hind leg"""
    duration = "1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Ranger', 'Sorcerer', 'Wizard')





#!/usr/bin/env python

"""Launch a system to interactively create a character."""

import logging
# logging.basicConfig(filename='character_creater.log', level=logging.DEBUG)
log = logging.getLogger(__name__)

import math
import numpy as np
import os
from random import randint
import subprocess

import npyscreen
import jinja2

from dungeonsheets import (character, race, dice, background, classes, weapons,
                           armor)


def read_version():
    version = open(os.path.join(os.path.dirname(__file__), '../VERSION')).read()
    version = version.replace('\n', '')
    return version


char_classes = {c.name: c for c in classes.available_classes}

races = {r.name: r for r in race.available_races}

backgrounds = {b.name: b for b in background.available_backgrounds}

all_weapons = (weapons.simple_melee_weapons + weapons.martial_melee_weapons +
               weapons.simple_ranged_weapons + weapons.martial_ranged_weapons)


class LinkedListForm(npyscreen.ActionForm):
    prev_page = None
    this_page = None
    next_page = None

    def __init__(self, formid, *args, **kwargs):
        self.this_page = formid
        super().__init__(*args, **kwargs)

    def to_next(self):
        self.parentApp.setNextForm(self.next_page)

    def to_prev(self):
        self.parentApp.setNextForm(self.prev_page)

    def add_next(self, next_name):
        new_next = self.parentApp.getForm(next_name)
        if self.next_page:
            current_next = self.parentApp.getForm(self.next_page)
            current_next.prev_page = next_name
        new_next.next_page = self.next_page
        new_next.prev_page = self.this_page
        self.next_page = next_name

    def add_prev(self, prev_name):
        new_prev = self.parentApp.getForm(prev_name)
        if self.prev_page:
            current_prev = self.parentApp.getForm(self.prev_page)
            current_prev.next_page = prev_name
        new_prev.prev_page = self.prev_page
        new_prev.next_page = self.this_page
        self.prev_page = prev_name

    def prune(self):
        if self.next_page:
            next_form = self.parentApp.getForm(self.next_page)
            next_form.prev_page = self.prev_page
        if self.prev_page:
            prev_form = self.parentApp.getForm(self.prev_page)
            prev_form.next_page = self.next_page
        self.parentApp.removeForm(self.this_page)


class App(npyscreen.NPSAppManaged):
    # STARTING_FORM = 'SKILLS'
    character = None
    n_classes = 1
    
    def save_character(self):
        # Save the file
        filename = self.getForm("SAVE").filename.value
        self.character.save(filename, template_file='empty_template.txt')
        # Create the PDF character sheet
        if self.getForm('SAVE').make_pdf.value:
            log.debug("Creating PDF")
            self.character.to_pdf(filename)
            subprocess.call(['makesheets', filename])

    def update_max_hp_roll(self):
        abil = self.getForm("ABILITIES")
        # Update max HP based on the class
        hit_dice = [dice.read_dice_str(d)
                    for d in self.character.hit_dice.split(' + ')]
        const = int((int(abil.constitution.value) - 10)/2)
        hp_max = f"{hit_dice[0].faces + self.character.level*const}"
        hds = {}
        for i, hd in enumerate(hit_dice):
            num = hd.num
            if i == 0:
                num -= 1
            if hd.faces not in hds:
                hds[hd.faces] = 0
            hds[hd.faces] += num
        for faces, num in hds.items():
            hp_max += f' + {num}d{faces}'
        abil.hp_roll_text.value = "Enter (or roll) your Max HP: " + hp_max
        abil.display()

    def reroll_max_hp(self):
        abil = self.getForm("ABILITIES")
        # Update max HP based on the class
        hit_dice = [dice.read_dice_str(d)
                    for d in self.character.hit_dice.split(' + ')]
        const = int((int(abil.constitution.value) - 10)/2)
        # Assume first hd given is from primary class
        hp_max = hit_dice[0].faces + const
        for i, hd in enumerate(hit_dice):
            num = hd.num
            if i == 0:
                num -= 1
            for d in range(num):
                hp_max += np.random.randint(low=1, high=hd.faces+1) + const
        abil.hp_max.value = str(hp_max)
        abil.display()
            
    def set_default_hp_max(self):
        abil = self.getForm("ABILITIES")
        # Update max HP based on the class
        hit_dice = [dice.read_dice_str(d)
                    for d in self.character.hit_dice.split(' + ')]
        const = int((int(abil.constitution.value) - 10)/2)
        # Assume first hd given is from primary class
        hp_max = hit_dice[0].faces + const
        for i, hd in enumerate(hit_dice):
            num = hd.num
            if i == 0:
                num -= 1
            for d in range(num):
                hp_max += math.ceil(hd.faces/2) + const
        log.debug("Updating max hp: %d", hp_max)
        abil.hp_max.value = str(hp_max)
        abil.display()
    
    def onStart(self):
        self.character = character.Character()
        self.character.class_list = []
        self.addForm("MAIN", BasicInfoForm, name="Basic Info:", formid='MAIN')
        self.addForm("RACE", RaceForm, name="Select your character's race:",
                     formid='RACE')
        self.addForm("CLASS1", CharacterClassForm, name="Select your character's primary class:",
                     formid='CLASS1')
        self.addForm("BACKGROUND", BackgroundForm, name="Choose background:",
                     formid='BACKGROUND')
        self.addForm("ALIGNMENT", AlignmentForm,
                     name="Select your character's alignment:",
                     formid='ALIGNMENT')
        self.addForm("ABILITIES", AbilityScoreForm,
                     name="Choose ability scores:", formid='ABILITIES')
        self.addForm("SKILLS", SkillForm, name="Choose skill proficiencies",
                     formid='SKILLS')
        self.addForm("WEAPONS", WeaponForm, name="Choose weapons",
                     formid='WEAPONS')
        self.addForm("ARMOR", ArmorForm, name="Choose armor",
                     formid='ARMOR')
        self.addForm("PERSONALITY", PersonalityForm, name="Describe your character",
                     formid='PERSONALITY')
        self.addForm("SAVE", SaveForm, name="Save character:", formid='SAVE')

        # Initialized the DoublyLinkedList
        forms = ['MAIN', 'RACE', 'CLASS1', 'BACKGROUND', 'ALIGNMENT',
                 'ABILITIES', 'SKILLS', 'WEAPONS', 'ARMOR', 'PERSONALITY',
                 'SAVE']
        for i in range(len(forms)-1):
            form = self.getForm(forms[i])
            form.add_next(forms[i+1])


class BasicInfoForm(LinkedListForm):
    def create(self):
        self.name = self.add(
            npyscreen.TitleText, name="Character Name:", use_two_lines=False)
        self.player_name = self.add(
            npyscreen.TitleText, name="Player Name:", use_two_lines=False)
    
    def on_ok(self):
        # Update the default filename
        name = self.name.value or "New Character"
        if name == '':
            filename = 'new_character.py'
        else:
            filename = f'{name.split(" ")[0].lower()}.py'
        save_form = self.parentApp.getForm('SAVE')
        save_form.filename.value = filename
        self.parentApp.character.name = self.name.value
        self.parentApp.character.player_name = self.player_name.value
        super().to_next()
    
    def on_cancel(self):
        raise KeyboardInterrupt

    
class RaceForm(LinkedListForm):
    def create(self):
        self.race = self.add(
            npyscreen.TitleSelectOne, name="Race:", values=tuple(races.keys()))
    
    def on_ok(self):
        if len(self.race.get_selected_objects()) >= 1:
            selected_race = self.race.get_selected_objects()[0]
            SelectedRace = races[selected_race]
            log.debug('Selected character race: %s', SelectedRace.name)
            self.parentApp.character.race = SelectedRace
            super().to_next()

    def on_cancel(self):
        super().to_prev()
        

class CharacterClassForm(LinkedListForm):
    class_num = 1

    def __init__(self, num=1, **kwargs):
        self.class_num = num
        self.class_options = list(char_classes.keys())
        super().__init__(**kwargs)

    @property
    def subclass_page(self):
        key = "SUBCLASS{:d}".format(self.class_num)
        if key in self.parentApp._Forms:
            return self.parentApp.getForm(key)
        else:
            return None

    @property
    def next_multiclass_page(self):
        key = "CLASS{:d}".format(self.class_num+1)
        if key in self.parentApp._Forms:
            return self.parentApp.getForm(key)
        else:
            return None

    def update_options(self):
        if len(self.parentApp.character.class_list) == 0:
            return
        else:
            self.class_options = list(char_classes.keys())
            for c in self.parentApp.character.class_list[:self.class_num-1]:
                self.class_options.remove(c.name)
            self.character_class.values = sorted(tuple(self.class_options))
            self.character_class.update()
        
    def create(self):
        if self.class_num > 1:
            self.add(npyscreen.FixedText, editable=False,
                     value="Current Classes: {}".format(
                         self.parentApp.character.name))
        if self.class_num == 1:
            t = 'Primary Class:'
        else:
            t = 'Class #{:d}:'.format(self.class_num)
        for c in self.parentApp.character.class_list:
            self.class_options.remove(c.name)
        if self.class_num == 1:
            self.multiclass = self.add(npyscreen.Checkbox, name="Add Multiclass?".format(self.class_num + 1), value=False)
        else:
            self.multiclass = self.add(npyscreen.Checkbox, name="Add Class #{:d}?".format(self.class_num + 1), value=False)
        self.level = self.add(
            npyscreen.TitleText, name='Level:', value="1", use_two_lines=False)
        self.character_class = self.add(
            npyscreen.TitleSelectOne, name=t, values=tuple(self.class_options))

    def add_multiclass_page(self):
        new_name = "CLASS{:d}".format(self.class_num + 1)
        new_form = self.parentApp.addForm(new_name,
                                          CharacterClassForm,
                                          name="Select your character's Class #{:d}:".format(self.class_num + 1),
                                          num=self.class_num+1,
                                          formid=new_name)
        self.add_next(new_name)
        return new_form

    def add_subclass_page(self, newclass, level):
        new_name = 'SUBCLASS{:d}'.format(self.class_num)
        new_form = self.parentApp.addForm(new_name,
                                          SubclassForm,
                                          name="Select your {:s} Subclass".format(newclass.name),
                                          newclass=newclass,
                                          level=level,
                                          num=self.class_num,
                                          formid=new_name)
        self.add_next(new_name)
        return new_form
        
    def on_ok(self):
        if len(self.character_class.get_selected_objects()) >= 1:
            selected_class = self.character_class.get_selected_objects()[0]
            selected_class = char_classes[selected_class]
            log.debug('Selected character class %s', selected_class.name)
            # replace later classes if we've backed up
            self.parentApp.character.class_list = self.parentApp.character.class_list[:self.class_num-1]
            self.parentApp.character.add_class(cls=selected_class,
                                               level=int(self.level.value),
                                               subclass=None)
            # add multiclass page if not exists yet
            if self.multiclass.value:
                if self.next_multiclass_page is None:
                    self.add_multiclass_page()
                else:
                    self.next_multiclass_page.update_options()
            else:
                # in case returned a page, prune any future multiclasses
                while self.next_page != 'BACKGROUND':
                    f = self.parentApp.getForm(self.next_page)
                    f.prune()
            if int(self.level.value) >= selected_class.subclass_select_level:
                if not self.subclass_page:
                    self.add_subclass_page(newclass=selected_class,
                                           level=int(self.level.value))
            else:
                if self.subclass_page is not None:
                    f = self.parentApp.getForm(self.next_page)
                    f.prune()
            super().to_next()
        
    def on_cancel(self):
        super().to_prev()


class SubclassForm(LinkedListForm):
    def __init__(self, newclass, level, num=1, **kwargs):
        self.class_num = num
        self.parent_class = newclass
        if len(newclass.subclasses_available) > 0:
            self.subclass_options = [sc.name for sc in newclass.subclasses_available]
        else:
            self.subclass_options = ('None',)
        self.level = level
        super().__init__(**kwargs)
    
    def create(self):
        self.subclass = self.add(
            npyscreen.TitleSelectOne, name="Subclass:",
            values=tuple(self.subclass_options))
        
    def on_ok(self):
        if len(self.subclass.get_selected_objects()) >= 1:
            sc = self.subclass.get_selected_objects()[0]
            if sc in [None, '', 'None']:
                sc = None
            self.parentApp.character.class_list = self.parentApp.character.class_list[:self.class_num-1]
            self.parentApp.character.add_class(cls=self.parent_class,
                                               level=self.level,
                                               subclass=sc)
            super().to_next()
            
    def on_cancel(self):
        super().to_prev()


class BackgroundForm(LinkedListForm):
    def create(self):
        self.background = self.add(
            npyscreen.TitleSelectOne,
            name="Background:", values=tuple(backgrounds.keys()))
    
    def on_ok(self):
        if len(self.background.get_selected_objects()) >= 1:
            selected_bg = self.background.get_selected_objects()[0]
            Background = backgrounds[selected_bg]
            self.parentApp.character.background = Background()
            # Update the languages based on background and race
            race_languages = self.parentApp.character.race.languages
            languages = Background.languages + race_languages
            self.parentApp.character.languages = ', '.join(languages)
            log.debug("Selected character background: %s", Background.name)
            super().to_next()
    
    def on_cancel(self):
        super().to_prev()


class AlignmentForm(LinkedListForm):
    """Choose your character's alignment."""
    alignments = ('Lawful good', 'Neutral good', 'Chaotic good',
                  'Lawful neutral', 'True neutral', 'Chaotic neutral',
                  'Lawful evil', 'Neutral evil', 'Chaotic evil', )

    def create(self):
        self.alignment = self.add(
            npyscreen.TitleSelectOne, name="Alignment:", values=self.alignments)
    
    def on_ok(self):
        if len(self.alignment.get_selected_objects()) >= 1:
            selected_alignment = self.alignment.get_selected_objects()[0]  # values[self.alignment.value]
            log.debug('Selected character alignment %s', selected_alignment)
            self.parentApp.character.alignment = selected_alignment
            # prep additions to abilities page
            self.parentApp.getForm('ABILITIES').prep()
            super().to_next()
    
    def on_cancel(self):
        super().to_prev()


class AbilityScoreForm(LinkedListForm):
    num_rolls = 0

    def roll_dice(self):
        """Get six ability scores that can then be assigned to abilities.""" 
        def roll_score():
            # Roll 4 dice and add the 3 highest
            rolls = (randint(1, 6) for i in range(4))
            score = sum(sorted(rolls)[-3:])
            return score
        scores = (roll_score() for i in range(6))
        return tuple(sorted(scores, reverse=True))

    def default_array(self):
        return (15, 14, 13, 12, 10, 8)

    def reroll(self, widget=None):
        self.num_rolls += 1
        new_scores = self.roll_dice()
        self.score_options.value = str(new_scores)[1:-1]
        self.score_options.update()
        self.reroll_button.value = False
        self.reroll_button.name = 'Reroll'
        self.reroll_button.update()
        self.default_button.value = False
        self.default_button.update()

    def set_default(self, widget=None):
        new_scores = self.default_array()
        self.score_options.value = str(new_scores)[1:-1]
        self.score_options.update()
        self.default_button.value = True
        self.default_button.update()

    def reroll_hp(self, widget=None):
        self.parentApp.reroll_max_hp()
        self.parentApp.update_max_hp_roll()
    
    def create(self):
        self.roll_text = self.add(npyscreen.FixedText, editable=False,
                                  value="Take the six rolls below and assign each one to an ability.")
        self.score_options = self.add(
            npyscreen.TitleFixedText, name="Rolls:", editable=False,
            value=str(self.default_array())[1:-1])
        self.default_button = self.add(npyscreen.MiniButtonPress,
                                       name="Use Default Rolls",
                                       when_pressed_function=self.set_default)
        self.reroll_button = self.add(npyscreen.MiniButtonPress,
                                      name="Reroll",
                                      when_pressed_function=self.reroll)

    def prep(self):
        attrs = ('strength', 'dexterity', 'constitution',
                 'intelligence', 'wisdom', 'charisma')
        self.class_text = self.add(npyscreen.FixedText, editable=False,
                                  value="Key stats for your primary class {:s} are listed with **".format(self.parentApp.character.primary_class.name))
        self.race_text = self.add(npyscreen.FixedText, editable=False,
                                  value="Do not add racial bonuses, they will be added for you as listed.")
        for attr in attrs:
            if attr in self.parentApp.character.primary_class.primary_abilities:
                name = '**' + attr
            else:
                name = '' + attr
            race_bonus = getattr(self.parentApp.character.race,
                                 f'{attr}_bonus')
            if race_bonus != 0:
                name += '({:+d})'.format(race_bonus)
            name += ':'
            new_fld = self.add(npyscreen.TitleText,
                               name=name,
                               begin_entry_at=24, value='10')
            setattr(self, attr, new_fld)
        self.hp_roll_text = self.add(npyscreen.FixedText, editable=False,
                                     value="")
        self.hp_reroll_buttom = self.add(npyscreen.MiniButtonPress,
                                         name="Reroll Max HP",
                                         when_pressed_function=self.reroll_hp)
        self.hp_max = self.add(npyscreen.TitleText, name="Max HP:")
        self.parentApp.update_max_hp_roll()
        self.parentApp.set_default_hp_max()

    def on_ok(self):
        attrs = ('strength', 'dexterity', 'constitution',
                 'intelligence', 'wisdom', 'charisma', 'hp_max')
        for attr in attrs:
            fld = getattr(self, attr)
            race_bonus = getattr(self.parentApp.character.race,
                                 f'{attr}_bonus', 0)
            try:
                val = int(float(fld.value))
            except ValueError:
                # Not an integer, so clear the field
                val = 10
            val += race_bonus
            log.debug("Setting %s to %s", attr, str(val))
            # Update the "character" with new values
            setattr(self.parentApp.character, attr, val)
        super().to_next()
    
    def on_cancel(self):
        super().to_prev()


class SkillForm(LinkedListForm):
    def while_editing(self):
        # Update the static skills for race and background
        bg_skills = self.parentApp.character.background.skill_proficiencies
        self.bg_skills.value = str(bg_skills)[1:-1].replace("'", "")
        race_skills = self.parentApp.character.race.skill_proficiencies
        self.race_skills.value = str(race_skills)[1:-1].replace("'", "")
        # Now set the available discretionary choices
        choices = (self.parentApp.character.primary_class.class_skill_choices +
                   self.parentApp.character.race.skill_choices +
                   self.parentApp.character.background.skill_choices)
        static_skills = bg_skills + race_skills
        choices = set([c for c in choices if c.lower() not in static_skills])
        self.skill_proficiencies.set_values(sorted(tuple(choices)))
        self.update_remaining()
    
    def update_remaining(self, widget=None):
        num_choices = (self.parentApp.character.primary_class.num_skill_choices +
                       self.parentApp.character.race.num_skill_choices +
                       self.parentApp.character.background.num_skill_choices)
        num_selected = len(self.skill_proficiencies.value)
        remaining = num_choices - num_selected
        log.debug(f'Remaining: {remaining}')
        self.remaining.value = str(remaining)
        self.display()
    
    def create(self):
        self.bg_skills = self.add(
            npyscreen.TitleText, name="Background:",
            value="", editable=False)
        self.race_skills = self.add(
            npyscreen.TitleText, name="Racial:",
            value="", editable=False)
        self.remaining = self.add(
            npyscreen.TitleText, name="Remaining:",
            value=0, editable=False)
        self.skill_proficiencies = self.add(
            npyscreen.TitleMultiSelect, name="Skill Proficiencies:",
            values=(),
            value_changed_callback=self.update_remaining)
    
    def on_ok(self):
        new_skills = self.skill_proficiencies.get_selected_objects()
        if new_skills is not None:
            new_skills = tuple(s.lower() for s in new_skills)
        else:
            new_skills = ()
        bg_skills = tuple(self.parentApp.character.background.skill_proficiencies)
        race_skills = tuple(self.parentApp.character.race.skill_proficiencies)
        all_skills = new_skills + bg_skills + race_skills
        self.parentApp.character.skill_proficiencies = all_skills
        self.parentApp.getForm('WEAPONS').update_options()
        log.debug(f"Skill proficiencies: {all_skills}")
        super().to_next()
    
    def on_cancel(self):
        super().to_prev()


class WeaponForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText, editable=False,
            value=("Please select your weapons."))
        self.remaining = self.add(
            npyscreen.TitleText, name="Remaining:",
            value=3, editable=False)
        self.weapons = self.add(
            npyscreen.TitleMultiSelect, name="Weapons:",
            values=tuple([wpn.name for wpn in all_weapons]),
            value_changed_callback=self.update_remaining)
    
    def on_ok(self):
        new_weapons = self.weapons.get_selected_objects()
        if new_weapons is not None:
            new_weapons = tuple(s.lower() for s in new_weapons)
        else:
            new_weapons = ()
        for wpn in new_weapons:
            self.parentApp.character.wield_weapon(wpn)
        log.debug(f"Weapons wielded: {new_weapons}")
        super().to_next()
    
    def update_remaining(self, widget=None):
        num_choices = 3
        num_selected = len(self.weapons.value)
        remaining = num_choices - num_selected
        log.debug(f'Remaining: {remaining}')
        self.remaining.value = str(remaining)
        self.display()
    
    def update_options(self):
        available_weapons = []
        for wpn in all_weapons:
            if self.parentApp.character.is_proficient(wpn()):
                available_weapons.append(wpn.name)
        self.weapons.values = available_weapons
        self.parentApp.getForm('ARMOR').update_options()
        self.display()
    
    def on_cancel(self):
        super().to_prev()


class ArmorForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText, editable=False,
            value=("Please select your armor."))
        self.shield = self.add(npyscreen.Checkbox, name="Shield? ",
                               value=False)
        self.armor = self.add(
            npyscreen.TitleSelectOne, name="Armor:",
            values=tuple([a.name for a in ([armor.NoArmor] + armor.all_armors)]))
        
    def on_ok(self):
        my_armor = self.armor.get_selected_objects()[0]
        if my_armor.lower() is not "no armor":
            self.parentApp.character.wear_armor(my_armor)
        if self.shield.value:
            self.parentApp.character.wield_shield('shield')
        super().to_next()
    
    def update_options(self):
        available_armors = [armor.NoArmor]
        proficiencies = self.parentApp.character.proficiencies_text.lower()
        if ('light armor' in proficiencies) or ('all armor' in proficiencies):
            available_armors.extend(armor.light_armors)
        if ('medium armor' in proficiencies) or ('all armor' in proficiencies):
            available_armors.extend(armor.medium_armors)
        if ('heavy armor' in proficiencies) or ('all armor' in proficiencies):
            available_armors.extend(armor.heavy_armors)
        self.armor.values = [a.name for a in available_armors]
        self.display()
    
    def on_cancel(self):
        super().to_prev()


class PersonalityForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText, editable=False,
            value=("Please describe your character's personality."))
        self.add(npyscreen.FixedText, editable=False,
                value=" ")
        self.pers_instr = self.add(npyscreen.FixedText, editable=False,
                value="Describe how your character behaves/interacts with others?")
        self.personality_traits = self.add(npyscreen.TitleText,
                                           name='Personality Traits: ',
                                           begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False,
                value=" ")
        self.ideal_instr = self.add(npyscreen.FixedText, editable=False,
                                    value="Describe what values does your character believes in?")
        self.ideals = self.add(
            npyscreen.TitleText, name='Ideals: ', begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False,
                value=" ")
        self.bond_instr = self.add(npyscreen.FixedText, editable=False,
                                   value="Describe your character's commitments or ongoing quests.")
        self.bonds = self.add(
            npyscreen.TitleText, name='Bonds: ', begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False,
                value=" ")
        self.flaw_instr = self.add(npyscreen.FixedText, editable=False,
                                   value="Describe your character's interesting flaws.")
        self.flaws = self.add(
            npyscreen.TitleText, name='Flaws: ', begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False,
                value=" ")
        self.feat_instr = self.add(npyscreen.FixedText, editable=False,
                                   value="Describe any other notable features or abilities.")
        self.features = self.add(
            npyscreen.TitleText, name='Features: ', begin_entry_at=24)
        
    def on_ok(self):
        if self.personality_traits.value:
            self.parentApp.character.personality_traits = self.personality_traits.value
        if self.ideals.value:
            self.parentApp.character.ideals = self.ideals.value
        if self.bonds.value:
            self.parentApp.character.bonds = self.bonds.value
        if self.flaws.value:
            self.parentApp.character.flaws = self.flaws.value
        if self.features.value:
            self.parentApp.character.features_and_traits = self.featurs.value
        super().to_next()
    
    def on_cancel(self):
        super().to_prev()


class SaveForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText, editable=False,
            value=("Your character will be saved in the file given below. "
                   "After saving, edit this file to finish your personality, "
                   "weapons, etc."))
        self.filename = self.add(
            npyscreen.TitleText, name='Filename:')
        self.make_pdf = self.add(npyscreen.Checkbox, name="Create PDF:",
                                 value=True)

    def on_ok(self):
        super().to_next()
    
    def on_cancel(self):
        super().to_prev()


def main():
    my_app = App()
    
    try:
        my_app.run()
    except KeyboardInterrupt:
        log.error("Aborted by user request")
    else:
        my_app.save_character()


if __name__ == '__main__':
    main()



from .features import Feature, create_feature

from .barbarian import *
from .bard import *
from .cleric import *
from .druid import *
from .fighter import *
from .monk import *
from .paladin import *
from .ranger import *
from .rogue import *
from .sorceror import *
from .warlock import *
from .wizard import *
from .races import *
from .backgrounds import *
from .feats import *
